#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import base64
import functools
import json
import logging
import sys
import threading
import time
import warnings
from collections import OrderedDict
from datetime import datetime

import requests

from .. import serializers, utils, errors, compat, readers, options
from ..accounts import BearerTokenAccount
from ..compat import Enum, six
from ..lib.monotonic import monotonic
from ..utils import to_str
from .core import LazyLoad, XMLRemoteModel, JSONRemoteModel
from .job import Job
from .readers import TunnelRecordReader, TunnelArrowReader
from .worker import WorkerDetail2, LOG_TYPES_MAPPING

logger = logging.getLogger(__name__)


_RESULT_LIMIT_HELPER_MSG = (
    'See https://pyodps.readthedocs.io/zh_CN/latest/base-sql.html#read-sql-exec-result '
    'for more information about limits on instance results.'
)


def _with_status_api_lock(func):
    @six.wraps(func)
    def wrapped(self, *args, **kw):
        with self._status_api_lock:
            return func(self, *args, **kw)

    return wrapped


class SpawnedInstanceReaderMixin(object):
    @property
    def schema(self):
        return self._schema

    @staticmethod
    def _read_instance_split(
        conn, download_id, start, count, idx,
        rest_client=None, project=None, instance_id=None, tunnel_endpoint=None,
        columns=None
    ):
        # read part data
        from ..tunnel import InstanceTunnel

        try:
            instance_tunnel = InstanceTunnel(
                client=rest_client, project=project, endpoint=tunnel_endpoint
            )
            session = instance_tunnel.create_download_session(
                instance=instance_id, download_id=download_id
            )
            with session.open_record_reader(start, count, columns=columns) as reader:
                conn.send((idx, reader.to_pandas(), True))
        except:
            conn.send((idx, sys.exc_info(), False))

    def _get_process_split_reader(self, columns=None):
        rest_client = self._parent._client
        project = self._parent.project.name
        tunnel_endpoint = self._parent.project._tunnel_endpoint
        instance_id = self._parent.id

        return functools.partial(
            self._read_instance_split,
            rest_client=rest_client,
            project=project,
            instance_id=instance_id,
            tunnel_endpoint=tunnel_endpoint,
            columns=columns or self._column_names,
        )


class InstanceRecordReader(SpawnedInstanceReaderMixin, TunnelRecordReader):
    def __init__(self, instance, download_session):
        super(InstanceRecordReader, self).__init__(
            instance, download_session
        )
        self._schema = download_session.schema


class InstanceArrowReader(SpawnedInstanceReaderMixin, TunnelArrowReader):
    def __init__(self, instance, download_session):
        super(InstanceArrowReader, self).__init__(
            instance, download_session
        )
        self._schema = download_session.schema


class Instance(LazyLoad):
    """
    Instance means that a ODPS task will sometimes run as an instance.

    ``status`` can reflect the current situation of a instance.
    ``is_terminated`` method indicates if the instance has finished.
    ``is_successful`` method indicates if the instance runs successfully.
    ``wait_for_success`` method will block the main process until the instance has finished.

    For a SQL instance, we can use open_reader to read the results.

    :Example:

    >>> instance = odps.execute_sql('select * from dual')  # this sql return the structured data
    >>> with instance.open_reader() as reader:
    >>>     # handle the record
    >>>
    >>> instance = odps.execute_sql('desc dual')  # this sql do not return structured data
    >>> with instance.open_reader() as reader:
    >>>    print(reader.raw)  # just return the raw result
    """

    __slots__ = (
        "_task_results",
        "_is_sync",
        "_instance_tunnel",
        "_id_thread_local",
        "_status_api_lock",
        "_logview_address",
        "_logview_address_time",
        "_logview_logged",
    )

    _download_id = utils.thread_local_attribute('_id_thread_local', lambda: None)

    def __init__(self, **kwargs):
        if 'task_results' in kwargs:
            kwargs['_task_results'] = kwargs.pop('task_results')
        super(Instance, self).__init__(**kwargs)

        try:
            del self._id_thread_local
        except AttributeError:
            pass

        if self._task_results is not None and len(self._task_results) > 0:
            self._is_sync = True
            self._status = Instance.Status.TERMINATED
        else:
            self._is_sync = False

        self._status_api_lock = threading.RLock()

        self._logview_address = None
        self._logview_address_time = None
        self._logview_logged = False

    @property
    def id(self):
        return self.name

    class Status(Enum):
        RUNNING = 'Running'
        SUSPENDED = 'Suspended'
        TERMINATED = 'Terminated'

    class InstanceStatus(XMLRemoteModel):
        _root = 'Instance'

        status = serializers.XMLNodeField('Status')

    class InstanceResult(XMLRemoteModel):

        class TaskResult(XMLRemoteModel):

            class Result(XMLRemoteModel):

                transform = serializers.XMLNodeAttributeField(attr='Transform')
                format = serializers.XMLNodeAttributeField(attr='Format')
                text = serializers.XMLNodeField('.', default='')

                def __str__(self):
                    if six.PY2:
                        text = utils.to_binary(self.text)
                    else:
                        text = self.text
                    if self.transform is not None and self.transform == 'Base64':
                        try:
                            return utils.to_str(base64.b64decode(text))
                        except TypeError:
                            return text
                    return text

                def __bytes__(self):
                    text = utils.to_binary(self.text)
                    if self.transform is not None and self.transform == 'Base64':
                        try:
                            return utils.to_binary(base64.b64decode(text))
                        except TypeError:
                            return text
                    return text

            type = serializers.XMLNodeAttributeField(attr='Type')
            name = serializers.XMLNodeField('Name')
            result = serializers.XMLNodeReferenceField(Result, 'Result')

        task_results = serializers.XMLNodesReferencesField(TaskResult, 'Tasks', 'Task')

    class Task(XMLRemoteModel):
        """
        Task stands for each task inside an instance.

        It has a name, a task type, the start to end time, and a running status.
        """

        name = serializers.XMLNodeField('Name')
        type = serializers.XMLNodeAttributeField(attr='Type')
        start_time = serializers.XMLNodeField('StartTime', parse_callback=utils.parse_rfc822)
        end_time = serializers.XMLNodeField('EndTime', parse_callback=utils.parse_rfc822)
        status = serializers.XMLNodeField(
            'Status', parse_callback=lambda s: Instance.Task.TaskStatus(s.upper()))
        histories = serializers.XMLNodesReferencesField('Instance.Task', 'Histories', 'History')

        class TaskStatus(Enum):
            WAITING = 'WAITING'
            RUNNING = 'RUNNING'
            SUCCESS = 'SUCCESS'
            FAILED = 'FAILED'
            SUSPENDED = 'SUSPENDED'
            CANCELLED = 'CANCELLED'

        class TaskProgress(XMLRemoteModel):
            """
            TaskProgress reprents for the progress of a task.

            A single TaskProgress may consist of several stages.

            :Example:

            >>> progress = instance.get_task_progress('task_name')
            >>> progress.get_stage_progress_formatted_string()
            2015-11-19 16:39:07 M1_Stg1_job0:0/0/1[0%]	R2_1_Stg1_job0:0/0/1[0%]
            """

            class StageProgress(XMLRemoteModel):

                name = serializers.XMLNodeAttributeField(attr='ID')
                backup_workers = serializers.XMLNodeField('BackupWorkers', parse_callback=int)
                terminated_workers = serializers.XMLNodeField('TerminatedWorkers', parse_callback=int)
                running_workers = serializers.XMLNodeField('RunningWorkers', parse_callback=int)
                total_workers = serializers.XMLNodeField('TotalWorkers', parse_callback=int)
                input_records = serializers.XMLNodeField('InputRecords', parse_callback=int)
                output_records = serializers.XMLNodeField('OutputRecords', parse_callback=int)
                finished_percentage = serializers.XMLNodeField('FinishedPercentage', parse_callback=int)

            stages = serializers.XMLNodesReferencesField(StageProgress, 'Stage')

            def get_stage_progress_formatted_string(self):
                buf = six.StringIO()

                buf.write(datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
                buf.write(' ')

                for stage in self.stages:
                    buf.write('{0}:{1}/{2}/{3}{4}[{5}%]\t'.format(
                        stage.name,
                        stage.running_workers,
                        stage.terminated_workers,
                        stage.total_workers,
                        '(+%s backups)' % stage.backup_workers if stage.backup_workers > 0 else '',
                        stage.finished_percentage
                    ))

                return buf.getvalue()

    class TaskInfo(serializers.XMLSerializableModel):
        _root = 'Instance'
        __slots__ = 'key', 'value'

        key = serializers.XMLNodeField('Key')
        value = serializers.XMLNodeField('Value')

    class TaskCost(object):
        __slots__ = 'cpu_cost', 'memory_cost', 'input_size'

        def __init__(self, cpu_cost=None, memory_cost=None, input_size=None):
            self.cpu_cost = cpu_cost
            self.memory_cost = memory_cost
            self.input_size = input_size

    class SQLCost(object):
        __slots__ = 'udf_num', 'complexity', 'input_size'

        def __init__(self, udf_num=None, complexity=None, input_size=None):
            self.udf_num = udf_num
            self.complexity = complexity
            self.input_size = input_size

    class DownloadSessionCreationError(errors.InternalServerError):
        pass

    name = serializers.XMLNodeField('Name')
    owner = serializers.XMLNodeField('Owner')
    start_time = serializers.XMLNodeField('StartTime', parse_callback=utils.parse_rfc822)
    end_time = serializers.XMLNodeField('EndTime', parse_callback=utils.parse_rfc822)
    _status = serializers.XMLNodeField('Status', parse_callback=lambda s: Instance.Status(s))
    _tasks = serializers.XMLNodesReferencesField(Task, 'Tasks', 'Task')

    class TaskSummary(dict):
        def __init__(self, *args, **kwargs):
            super(Instance.TaskSummary, self).__init__(*args, **kwargs)
            self.summary_text, self.json_summary = None, None

    class AnonymousSubmitInstance(XMLRemoteModel):
        _root = 'Instance'
        job = serializers.XMLNodeReferenceField(Job, 'Job')

    class InstanceQueueingInfo(JSONRemoteModel):
        __slots__ = '_instance',

        class Status(Enum):
            RUNNING = 'Running'
            SUSPENDED = 'Suspended'
            TERMINATED = 'Terminated'
            UNKNOWN = 'Unknown'

        _properties = serializers.JSONRawField()  # hold the raw dict
        instance_id = serializers.JSONNodeField('instanceId')
        priority = serializers.JSONNodeField('instancePriority')
        progress = serializers.JSONNodeField('instanceProcess')
        job_name = serializers.JSONNodeField('jobName')
        project = serializers.JSONNodeField('projectName')
        skynet_id = serializers.JSONNodeField('skynetId')
        start_time = serializers.JSONNodeField('startTime', parse_callback=utils.strptime_with_tz)
        task_type = serializers.JSONNodeField('taskType')
        task_name = serializers.JSONNodeField('taskName')
        user_account = serializers.JSONNodeField('userAccount')
        status = serializers.JSONNodeField('status', parse_callback=Status)

        @property
        def instance(self):
            if hasattr(self, '_instance') and self._instance:
                return self._instance

            from .projects import Projects
            self._instance = Projects(client=self._client)[self.project].instances[self.instance_id]
            return self._instance

        def __getattr__(self, item):
            item = utils.underline_to_camel(item)

            if item in self._properties:
                return self._properties[item]

            return super(Instance.InstanceQueueingInfo, self).__getattr__(item)

    def reload(self):
        resp = self._client.get(self.resource())

        self.owner = resp.headers.get('x-odps-owner')
        self.start_time = utils.parse_rfc822(resp.headers.get('x-odps-start-time'))
        end_time_header = 'x-odps-end-time'
        if end_time_header in resp.headers and \
                len(resp.headers[end_time_header].strip()) > 0:
            self.end_time = utils.parse_rfc822(resp.headers.get(end_time_header))

        self.parse(self._client, resp, obj=self)
        # remember not to set `_loaded = True`

    def stop(self):
        """
        Stop this instance.

        :return: None
        """

        instance_status = Instance.InstanceStatus(status='Terminated')
        xml_content = instance_status.serialize()

        headers = {'Content-Type': 'application/xml'}
        self._client.put(self.resource(), xml_content, headers=headers)

    @_with_status_api_lock
    def get_task_results_without_format(self, timeout=None):
        if self._is_sync:
            return self._task_results

        resp = self._client.get(self.resource(), action="result", timeout=timeout)

        instance_result = Instance.InstanceResult.parse(self._client, resp)
        return OrderedDict([(r.name, r.result) for r in instance_result.task_results])

    @_with_status_api_lock
    def get_task_results(self, timeout=None):
        """
        Get all the task results.

        :return: a dict which key is task name, and value is the task result as string
        :rtype: dict
        """

        results = self.get_task_results_without_format(timeout=timeout)
        if options.tunnel.string_as_binary:
            return OrderedDict([(k, bytes(result)) for k, result in six.iteritems(results)])
        else:
            return OrderedDict([(k, str(result)) for k, result in six.iteritems(results)])

    @_with_status_api_lock
    def get_task_result(self, task_name, timeout=None):
        """
        Get a single task result.

        :param task_name: task name
        :return: task result
        :rtype: str
        """
        return self.get_task_results(timeout=timeout).get(task_name)

    @_with_status_api_lock
    def get_task_summary(self, task_name):
        """
        Get a task's summary, mostly used for MapReduce.

        :param task_name: task name
        :return: summary as a dict parsed from JSON
        :rtype: dict
        """

        params = {'taskname': task_name}
        resp = self._client.get(
            self.resource(), action='instancesummary', params=params
        )

        map_reduce = resp.json().get('Instance')
        if map_reduce:
            json_summary = map_reduce.get('JsonSummary')
            if json_summary:
                summary = Instance.TaskSummary(json.loads(json_summary))
                summary.summary_text = map_reduce.get('Summary')
                summary.json_summary = json_summary

                return summary

    @_with_status_api_lock
    def get_task_statuses(self):
        """
        Get all tasks' statuses

        :return: a dict which key is the task name and value is the :class:`odps.models.Instance.Task` object
        :rtype: dict
        """
        resp = self._client.get(self.resource(), action="taskstatus")
        self.parse(self._client, resp, obj=self)

        return dict([(task.name, task) for task in self._tasks])

    @_with_status_api_lock
    def get_task_names(self):
        """
        Get names of all tasks

        :return: task names
        :rtype: list
        """

        return compat.lkeys(self.get_task_statuses())

    def get_task_cost(self, task_name):
        """
        Get task cost

        :param task_name: name of the task
        :return: task cost
        :rtype: Instance.TaskCost

        :Example:

        >>> cost = instance.get_task_cost(instance.get_task_names()[0])
        >>> cost.cpu_cost
        200
        >>> cost.memory_cost
        4096
        >>> cost.input_size
        0
        """
        summary = self.get_task_summary(task_name)
        if summary is None:
            return None

        if 'Cost' in summary:
            task_cost = summary['Cost']

            cpu_cost = task_cost.get('CPU')
            memory = task_cost.get('Memory')
            input_size = task_cost.get('Input')

            return Instance.TaskCost(cpu_cost, memory, input_size)

    def get_task_info(self, task_name, key):
        """
        Get task related information.

        :param task_name: name of the task
        :param key: key of the information item
        :return: a string of the task information
        """
        actions = ["info"]
        params = OrderedDict([('taskname', task_name), ('key', key)])

        resp = self._client.get(self.resource(), actions=actions, params=params)
        return resp.content.decode()

    def put_task_info(self, task_name, key, value, check_location=False):
        """
        Put information into a task.

        :param task_name: name of the task
        :param key: key of the information item
        :param value: value of the information item
        """
        actions = ["info"]
        params = {'taskname': task_name}
        headers = {'Content-Type': 'application/xml'}
        body = self.TaskInfo(key=key, value=value).serialize()

        resp = self._client.put(self.resource(), actions=actions, params=params, headers=headers, data=body)

        location = resp.headers.get('Location')
        if check_location and (location is None or len(location) == 0):
            raise errors.ODPSError('Invalid response, Location header required.')
        return resp.content.decode()

    def get_task_quota(self, task_name):
        """
        Get queueing info of the task.
        Note that time between two calls should larger than 30 seconds, otherwise empty dict is returned.

        :param task_name: name of the task
        :return: quota info in dict format
        """
        actions = ['instancequota']
        params = {'taskname': task_name}
        resp = self._client.get(self.resource(), actions=actions, params=params)
        return json.loads(resp.text)

    def get_sql_task_cost(self):
        """
        Get cost information of the sql cost task, including input data size,
        number of UDF, Complexity of the sql task.

        NOTE that DO NOT use this function directly as it cannot be applied to
        instances returned from SQL. Use ``o.execute_sql_cost`` instead.

        :return: cost info in dict format
        """
        resp = self.get_task_result(self.get_task_names()[0])
        cost = json.loads(resp)
        sql_cost = cost['Cost']['SQL']

        udf_num = sql_cost.get('UDF')
        complexity = sql_cost.get('Complexity')
        input_size = sql_cost.get('Input')
        return Instance.SQLCost(udf_num, complexity, input_size)

    @property
    @_with_status_api_lock
    def status(self):
        if self._status != Instance.Status.TERMINATED:
            self.reload()

        return self._status

    def is_terminated(self, retry=False):
        """
        If this instance has finished or not.

        :return: True if finished else False
        :rtype: bool
        """
        return utils.call_with_retry(
            lambda: self.status == Instance.Status.TERMINATED,
            retry_times=options.retry_times if retry else 0,
            exc_type=(errors.InternalServerError, errors.RequestTimeTooSkewed),
        )

    def is_running(self, retry=False):
        """
        If this instance is still running.

        :return: True if still running else False
        :rtype: bool
        """
        return utils.call_with_retry(
            lambda: self.status == Instance.Status.RUNNING,
            retry_times=options.retry_times if retry else 0,
            exc_type=(errors.InternalServerError, errors.RequestTimeTooSkewed),
        )

    def is_successful(self, retry=False):
        """
        If the instance runs successfully.

        :return: True if successful else False
        :rtype: bool
        """

        if not self.is_terminated(retry=retry):
            return False

        def _get_successful():
            statuses = self.get_task_statuses()
            return all(
                task.status == Instance.Task.TaskStatus.SUCCESS
                for task in statuses.values()
            )

        return utils.call_with_retry(
            _get_successful,
            retry_times=options.retry_times if retry else 0,
            exc_type=(errors.InternalServerError, errors.RequestTimeTooSkewed),
        )

    @property
    def is_sync(self):
        return self._is_sync

    def get_all_task_progresses(self):
        return {
            task_name: self.get_task_progress(task_name)
            for task_name in self.get_task_names()
        }

    def wait_for_completion(self, interval=1, timeout=None, max_interval=None):
        """
        Wait for the instance to complete, and neglect the consequence.

        :param interval: time interval to check
        :param max_interval: if specified, next check interval will be
            multiplied by 2 till max_interval is reached.
        :param timeout: time
        :return: None
        """

        start_time = monotonic()
        progress_time = monotonic()
        last_progress = 0
        while not self.is_terminated(retry=True):
            try:
                time.sleep(interval)
                check_time = monotonic()
                if max_interval is not None:
                    interval = min(interval * 2, max_interval)
                if timeout is not None and check_time - start_time > timeout:
                    raise errors.WaitTimeoutError(
                        "Wait completion of instance %s timed out" % self.id, instance_id=self.id
                    )

                if logger.getEffectiveLevel() <= logging.INFO:
                    try:
                        task_progresses = self.get_all_task_progresses()
                        total_progress = sum(
                            stage.finished_percentage
                            for progress in task_progresses.values()
                            for stage in progress.stages
                        )
                        if check_time - start_time >= options.progress_time_interval and (
                            total_progress - last_progress >= options.progress_percentage_gap
                            or check_time - progress_time >= options.progress_time_interval
                        ):
                            if not self._logview_logged:
                                self._logview_logged = True
                                logger.info(
                                    "Instance ID: %s\n  Log view: %s", self.id, self.get_logview_address()
                                )

                            output_parts = [str(self.id)] + [
                                progress.get_stage_progress_formatted_string()
                                for progress in task_progresses.values()
                            ]
                            if len(output_parts) > 1:
                                logger.info(" ".join(output_parts))
                            last_progress = total_progress
                            progress_time = check_time
                    except:  # pragma: no cover
                        # make sure progress display does not affect execution
                        pass
            except KeyboardInterrupt:
                break

    def wait_for_success(self, interval=1, timeout=None, max_interval=None):
        """
        Wait for instance to complete, and check if the instance is successful.

        :param interval: time interval to check
        :param max_interval: if specified, next check interval will be
            multiplied by 2 till max_interval is reached.
        :param timeout: time
        :return: None
        :raise: :class:`odps.errors.ODPSError` if the instance failed
        """

        self.wait_for_completion(interval=interval, max_interval=max_interval, timeout=timeout)

        if not self.is_successful(retry=True):
            for task_name, task in six.iteritems(self.get_task_statuses()):
                exc = None
                if task.status == Instance.Task.TaskStatus.FAILED:
                    exc = errors.parse_instance_error(self.get_task_result(task_name))
                elif task.status != Instance.Task.TaskStatus.SUCCESS:
                    exc = errors.ODPSError('%s, status=%s' % (task_name, task.status.value))
                if exc:
                    exc.instance_id = self.id
                    raise exc

    @_with_status_api_lock
    def get_task_progress(self, task_name):
        """
        Get task's current progress

        :param task_name: task_name
        :return: the task's progress
        :rtype: :class:`odps.models.Instance.Task.TaskProgress`
        """

        params = {'instanceprogress': task_name, 'taskname': task_name}

        resp = self._client.get(self.resource(), params=params)
        return Instance.Task.TaskProgress.parse(self._client, resp)

    @_with_status_api_lock
    def get_task_detail(self, task_name):
        """
        Get task's detail

        :param task_name: task name
        :return: the task's detail
        :rtype: list or dict according to the JSON
        """
        def _get_detail():
            from ..compat import json  # fix object_pairs_hook parameter for Py2.6

            params = {'taskname': task_name}
            resp = self._client.get(
                self.resource(), action='instancedetail', params=params
            )
            return json.loads(
                resp.content.decode() if six.PY3 else resp.content,
                object_pairs_hook=OrderedDict
            )

        result = _get_detail()
        if not result:
            # todo: this is a workaround for the bug that get_task_detail returns nothing.
            return self.get_task_detail2(task_name)
        else:
            return result

    @_with_status_api_lock
    def get_task_detail2(self, task_name):
        """
        Get task's detail v2

        :param task_name: task name
        :return: the task's detail
        :rtype: list or dict according to the JSON
        """
        params = {'taskname': task_name}

        resp = self._client.get(self.resource(), action='detail', params=params)
        res = resp.content.decode() if six.PY3 else resp.content
        try:
            return json.loads(res, object_pairs_hook=OrderedDict)
        except ValueError:
            return res

    @_with_status_api_lock
    def get_task_workers(self, task_name=None, json_obj=None):
        """
        Get workers from task
        :param task_name: task name
        :param json_obj: json object parsed from get_task_detail2
        :return: list of workers

        .. seealso:: :class:`odps.models.Worker`
        """
        if task_name is None and json_obj is None:
            raise ValueError('Either task_name or json_obj should be provided')

        if json_obj is None:
            json_obj = self.get_task_detail2(task_name)
        return WorkerDetail2.extract_from_json(json_obj, client=self._client, parent=self)

    @_with_status_api_lock
    def get_worker_log(self, log_id, log_type, size=0):
        """
        Get logs from worker.

        :param log_id: id of log, can be retrieved from details.
        :param log_type: type of logs. Possible log types contains {log_types}
        :param size: length of the log to retrieve
        :return: log content
        """
        params = OrderedDict([('log', ''), ('id', log_id)])
        if log_type is not None:
            log_type = log_type.lower()
            if log_type not in LOG_TYPES_MAPPING:
                raise ValueError('log_type should choose a value in ' +
                                 ' '.join(six.iterkeys(LOG_TYPES_MAPPING)))
            params['logtype'] = LOG_TYPES_MAPPING[log_type]
        if size > 0:
            params['size'] = str(size)
        resp = self._client.get(self.resource(), params=params)
        return resp.text
    get_worker_log.__doc__ = get_worker_log.__doc__.format(log_types=', '.join(sorted(six.iterkeys(LOG_TYPES_MAPPING))))

    @_with_status_api_lock
    def get_logview_address(self, hours=None):
        """
        Get logview address of the instance object by hours.

        :param hours:
        :return: logview address
        :rtype: str
        """
        if (
            self._logview_address is not None
            and monotonic() - self._logview_address_time < 600
        ):
            return self._logview_address

        project = self.project
        if isinstance(project.odps.account, BearerTokenAccount):
            token = to_str(project.odps.account.token)
        else:
            hours = hours or options.logview_hours
            policy = {
                'Statement': [{
                    'Action': ['odps:Read'],
                    'Effect': 'Allow',
                    'Resource': 'acs:odps:*:projects/%s/instances/%s' % (project.name, self.id)
                }],
                'Version': '1',
            }
            token = self.project.generate_auth_token(policy, "bearer", hours)

        link = (
            self.project.odps.logview_host
            + "/logview/?h="
            + self._client.endpoint
            + "&p="
            + project.name
            + "&i="
            + self.id
            + "&token="
            + token
        )
        self._logview_address = link
        self._logview_address_time = monotonic()
        return link

    def __str__(self):
        return self.id

    def _get_job(self):
        url = self.resource()
        resp = self._client.get(url, action='source')

        job = Job.parse(self._client, resp, parent=self)
        return job

    def get_tasks(self):
        return self.tasks

    @property
    def tasks(self):
        job = self._get_job()
        return job.tasks

    @property
    def priority(self):
        job = self._get_job()
        return job.priority

    def _get_queueing_info(self):
        url = self.resource()
        resp = self._client.get(url, action='cached')
        return Instance.InstanceQueueingInfo.parse(
            self._client, resp, parent=self.project.instance_queueing_infos), resp

    def get_queueing_info(self):
        info, _ = self._get_queueing_info()
        return info

    def _create_instance_tunnel(self, endpoint=None):
        if self._instance_tunnel is not None:
            return self._instance_tunnel

        from ..tunnel import InstanceTunnel

        self._instance_tunnel = InstanceTunnel(
            client=self._client, project=self.project,
            endpoint=endpoint or self.project._tunnel_endpoint
        )
        return self._instance_tunnel

    @utils.survey
    def _open_result_reader(self, schema=None, task_name=None, timeout=None, **_):
        if not self.is_successful(retry=True):
            raise errors.ODPSError(
                'Cannot open reader, instance(%s) may fail or has not finished yet' % self.id)

        sql_tasks = dict([(name, task) for name, task in six.iteritems(self.get_task_statuses())
                          if task.type.lower() == 'sql'])
        if len(sql_tasks) > 1:
            if task_name is None:
                raise errors.ODPSError(
                    'Cannot open reader, job has more than one sql tasks, please specify one')
            elif task_name not in sql_tasks:
                raise errors.ODPSError(
                    'Cannot open reader, unknown task name: %s' % task_name)
        elif len(sql_tasks) == 1:
            task_name = list(sql_tasks)[0]
        else:
            raise errors.ODPSError('Cannot open reader, job has no sql task')

        result = self.get_task_result(task_name, timeout=timeout)
        reader = readers.CsvRecordReader(schema, result)
        if options.result_reader_create_callback:
            options.result_reader_create_callback(reader)
        return reader

    def _open_tunnel_reader(self, **kw):
        from ..tunnel.instancetunnel import InstanceDownloadSession

        reopen = kw.pop('reopen', False)
        endpoint = kw.pop('endpoint', None)
        arrow = kw.pop("arrow", False)
        tunnel = self._create_instance_tunnel(endpoint=endpoint)
        download_id = self._download_id if not reopen else None

        try:
            download_session = tunnel.create_download_session(
                instance=self, download_id=download_id, **kw
            )
            if download_id and download_session.status != InstanceDownloadSession.Status.Normal:
                download_session = tunnel.create_download_session(instance=self, **kw)
        except errors.InternalServerError:
            e, tb = sys.exc_info()[1:]
            e.__class__ = Instance.DownloadSessionCreationError
            six.reraise(Instance.DownloadSessionCreationError, e, tb)

        self._download_id = download_session.id

        if arrow:
            return InstanceArrowReader(self, download_session)
        else:
            return InstanceRecordReader(self, download_session)

    def open_reader(self, *args, **kwargs):
        """
        Open the reader to read records from the result of the instance. If `tunnel` is `True`,
        instance tunnel will be used. Otherwise conventional routine will be used. If instance tunnel
        is not available and `tunnel` is not specified, the method will fall back to the
        conventional routine.
        Note that the number of records returned is limited unless `options.limited_instance_tunnel`
        is set to `True` or `limit=True` is configured under instance tunnel mode. Otherwise
        the number of records returned is always limited.

        :param tunnel: if true, use instance tunnel to read from the instance.
                       if false, use conventional routine.
                       if absent, `options.tunnel.use_instance_tunnel` will be used and automatic fallback
                       is enabled.
        :param bool limit: if True, enable the limitation
        :param bool reopen: the reader will reuse last one, reopen is true means open a new reader.
        :param endpoint: the tunnel service URL
        :param compress_option: compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :param compress_algo: compression algorithm, work when ``compress_option`` is not provided,
                              can be ``zlib``, ``snappy``
        :param compress_level: used for ``zlib``, work when ``compress_option`` is not provided
        :param compress_strategy: used for ``zlib``, work when ``compress_option`` is not provided
        :return: reader, ``count`` means the full size, ``status`` means the tunnel status

        :Example:

        >>> with instance.open_reader() as reader:
        >>>     count = reader.count  # How many records of a table or its partition
        >>>     for record in reader[0: count]:
        >>>         # read all data, actually better to split into reading for many times
        """
        use_tunnel = kwargs.get('use_tunnel', kwargs.get('tunnel'))
        auto_fallback_result = use_tunnel is None

        timeout = kwargs.pop("timeout", None)
        if use_tunnel is None:
            use_tunnel = options.tunnel.use_instance_tunnel
            if use_tunnel:
                timeout = timeout if timeout is not None else options.tunnel.legacy_fallback_timeout
        kwargs["timeout"] = timeout

        result_fallback_errors = (errors.InvalidProjectTable, errors.InvalidArgument)
        if use_tunnel:
            # for compatibility
            if 'limit_enabled' in kwargs:
                kwargs['limit'] = kwargs['limit_enabled']
                del kwargs['limit_enabled']

            if 'limit' not in kwargs:
                kwargs['limit'] = options.tunnel.limit_instance_tunnel

            auto_fallback_protection = False
            if kwargs['limit'] is None:
                kwargs['limit'] = False
                auto_fallback_protection = True

            try:
                return self._open_tunnel_reader(**kwargs)
            except result_fallback_errors:
                # service version too low to support instance tunnel.
                if not auto_fallback_result:
                    raise
                if not kwargs.get('limit'):
                    warnings.warn('Instance tunnel not supported, will fallback to '
                                  'conventional ways. 10000 records will be limited. '
                                  + _RESULT_LIMIT_HELPER_MSG)
            except requests.Timeout:
                # tunnel creation timed out, which might be caused by too many files
                # on the service.
                if not auto_fallback_result:
                    raise
                if not kwargs.get('limit'):
                    warnings.warn(
                        'Instance tunnel timed out, will fallback to conventional ways. '
                        '10000 records will be limited. You may try merging small files '
                        'on your source table. ' + _RESULT_LIMIT_HELPER_MSG)
            except (Instance.DownloadSessionCreationError, errors.InstanceTypeNotSupported):
                # this is for DDL sql instances such as `show partitions` which raises
                # InternalServerError when creating download sessions.
                if not auto_fallback_result:
                    raise
            except errors.NoPermission:
                # project is protected
                if not auto_fallback_protection:
                    raise
                if not kwargs.get('limit'):
                    warnings.warn('Project under protection, 10000 records will be limited.'
                                  + _RESULT_LIMIT_HELPER_MSG)
                    kwargs['limit'] = True
                    return self.open_reader(*args, **kwargs)

        return self._open_result_reader(*args, **kwargs)



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import sys
import json
from collections import namedtuple, OrderedDict

from ... import serializers, DataFrame
from ...compat import six
from ..utils import camel_to_underline, parse_hist_repr
from ..algolib.loader import load_defined_algorithms

INF_PATTERN = re.compile(r'(: *)(inf|-inf)( *,| *\})')
load_defined_algorithms(sys.modules[__name__], 'metrics')


def replace_json_infs(s):
    def repl(m):
        mid = m.group(2)
        if mid == 'inf':
            mid = 'Infinity'
        elif mid == '-inf':
            mid = '-Infinity'
        return m.group(1) + mid + m.group(3)

    return INF_PATTERN.sub(repl, s)


def get_confusion_matrix_result(expr, odps):
    records = list(DataFrame(odps.get_table(expr.tables[0])).execute())
    # skip the first row
    col_data = json.loads(records[0][0])
    col_data = map(lambda p: p[1], sorted(six.iteritems(col_data), key=lambda a: int(a[0][1:])))

    mat = [list(rec[2:]) for rec in records[1:]]
    try:
        import numpy as np
        mat = np.matrix(mat)
    except ImportError:
        pass
    return mat, col_data

RocResult = namedtuple('RocResult', 'thresh tp fn tn fp')


def get_roc_result(expr, odps):
    mat = [list(rec) for rec in DataFrame(odps.get_table(expr.tables[0])).execute()]
    thresh, tp, fn, tn, fp = [[r[ridx] for r in mat] for ridx in range(0, 5)]
    try:
        import numpy as np
        thresh, tp, fn, tn, fp = [np.asarray(l) for l in (thresh, tp, fn, tn, fp)]
    except ImportError:
        pass

    return RocResult(thresh, tp, fn, tn, fp)


def get_regression_eval_result(expr, odps):
    records = list(DataFrame(odps.get_table(expr.tables.index)).execute())
    indices = dict((camel_to_underline(k), v) for k, v in six.iteritems(json.loads(replace_json_infs(records[0].values[0]))))
    records = list(DataFrame(odps.get_table(expr.tables.residual)).execute())
    indices['hist'] = parse_hist_repr(records[0].values[1])
    return indices


class MultiClassEvaluationResult(serializers.JSONSerializableModel):
    class LabelMeasure(serializers.JSONSerializableModel):
        accuracy = serializers.JSONNodeField('Accuracy')
        f1 = serializers.JSONNodeField('F1')
        false_discovery_rate = serializers.JSONNodeField('FalseDiscoveryRate')
        kappa = serializers.JSONNodeField('Kappa')
        npv = serializers.JSONNodeField('NegativePredictiveValue')
        precision = serializers.JSONNodeField('Precision')
        sensitivity = serializers.JSONNodeField('Sensitivity')
        specificity = serializers.JSONNodeField('Specificity')
        fn = serializers.JSONNodeField('FalseNegative')
        fnr = serializers.JSONNodeField('FalseNegativeRate')
        fp = serializers.JSONNodeField('FalsePositive')
        fpr = serializers.JSONNodeField('FalsePositiveRate')
        tn = serializers.JSONNodeField('TrueNegative')
        tp = serializers.JSONNodeField('TruePositive')

        @property
        def tpr(self):
            return self.sensitivity

        @tpr.setter
        def tpr(self, value):
            self.sensitivity = value

        @property
        def fpr(self):
            return self.specificity

        @fpr.setter
        def fpr(self, value):
            self.specificity = value

    class OverallMeasures(serializers.JSONSerializableModel):
        accuracy = serializers.JSONNodeField('Accuracy')
        kappa = serializers.JSONNodeField('Kappa')
        frequency_micro = serializers.JSONNodeReferenceField('MultiClassEvaluationResult.LabelMeasure',
                                                             'LabelFrequencyBasedMicro')
        macro_average = serializers.JSONNodeReferenceField('MultiClassEvaluationResult.LabelMeasure',
                                                           'MacroAveraged')
        micro_average = serializers.JSONNodeReferenceField('MultiClassEvaluationResult.LabelMeasure',
                                                           'MicroAveraged')

    actual_label_counts = serializers.JSONNodeField('ActualLabelFrequencyList')
    actual_label_ratios = serializers.JSONNodeField('ActualLabelProportionList')

    predicted_label_counts = serializers.JSONNodeField('PredictedLabelFrequencyList')
    predicted_label_ratios = serializers.JSONNodeField('PredictedLabelProportionList')

    confusion_matrix = serializers.JSONNodeField('ConfusionMatrix', type='ndarray')
    ratio_matrix = serializers.JSONNodeField('ProportionMatrix', type='ndarray')

    labels = serializers.JSONNodeField('LabelList')
    label_measures = serializers.JSONNodesReferencesField(LabelMeasure, 'LabelMeasureList')


def get_multi_class_eval_result(expr, odps):
    rec = DataFrame(odps.get_table(expr.tables[0])).execute()[0]
    return MultiClassEvaluationResult.parse(rec[0])


class BinaryClassEvaluationResult(object):
    def __init__(self):
        self.records = []
        self.positive = 0
        self.negative = 0

    @staticmethod
    def _to_numpy(v):
        try:
            import numpy as np
            return np.asarray(v)
        except ImportError:
            return v

    @property
    def thresh(self):
        return self._to_numpy([rec['data_range'][1] for rec in self.records])

    @property
    def fp(self):
        return self._to_numpy([self.negative * rec['fpr'] for rec in self.records])

    @property
    def tn(self):
        return self._to_numpy([self.negative * (1 - rec['fpr']) for rec in self.records])

    @property
    def tp(self):
        return self._to_numpy([self.positive * rec['recall'] for rec in self.records])

    @property
    def fn(self):
        return self._to_numpy([self.positive * (1 - rec['recall']) for rec in self.records])


def get_binary_class_eval_result(expr, odps):
    result = BinaryClassEvaluationResult()

    metric_recs = DataFrame(odps.get_table(expr.tables.metric)).execute()

    targets = {
        'Total Samples': 'size',
        'Positive Samples': 'positive',
        'Negative Samples': 'negative',
        'AUC': 'auc',
        'KS': 'ks',
        'F1 Score': 'f1',
    }
    for rec in metric_recs:
        if rec['name'] not in targets:
            continue
        setattr(result, targets[rec['name']], rec['value'])

    detail_recs = DataFrame(odps.get_table(expr.tables.detail)).execute()
    new_recs = []
    for rec in detail_recs:
        if rec['total'] == 0:
            continue
        new_rec = OrderedDict(rec.iteritems())
        new_rec['data_range'] = tuple(float(v) for v in rec['data_range'].strip('[()]').split(','))
        new_recs.append(new_rec)
    result.records = new_recs
    return result


def get_clustering_eval_result(expr, odps):
    return json.loads(DataFrame(odps.get_table(expr.tables[0])).execute()[0][0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from .. import options, serializers, utils
from ..compat import six, quote_plus
from .cache import cache, del_cache


class XMLRemoteModel(serializers.XMLSerializableModel):
    __slots__ = '_parent', '_client', '_schema_name'

    def __init__(self, **kwargs):
        if 'parent' in kwargs:
            kwargs['_parent'] = kwargs.pop('parent')
        if 'client' in kwargs:
            kwargs['_client'] = kwargs.pop('client')

        self._schema_name = utils.notset

        if not frozenset(kwargs).issubset(self.__slots__):
            unexpected = sorted(set(kwargs) - set(self.__slots__))
            raise TypeError("%s() meet illegal arguments (%s)" % (
                type(self).__name__, ', '.join(unexpected)))
        super(XMLRemoteModel, self).__init__(**kwargs)

    @classmethod
    def parse(cls, client, response, obj=None, **kw):
        kw['_client'] = client
        return super(XMLRemoteModel, cls).parse(response, obj=obj, **kw)


class AbstractXMLRemoteModel(XMLRemoteModel):
    __slots__ = '_type_indicator',


class JSONRemoteModel(serializers.JSONSerializableModel):
    __slots__ = '_parent', '_client'

    def __init__(self, **kwargs):
        if 'parent' in kwargs:
            kwargs['_parent'] = kwargs.pop('parent')
        if 'client' in kwargs:
            kwargs['_client'] = kwargs.pop('client')
        if not frozenset(kwargs).issubset(self.__slots__):
            unexpected = sorted(set(kwargs) - set(self.__slots__))
            raise TypeError("%s() meet illegal arguments (%s)" % (
                type(self).__name__, ', '.join(unexpected)))
        super(JSONRemoteModel, self).__init__(**kwargs)

    @classmethod
    def parse(cls, client, response, obj=None, **kw):
        kw['_client'] = client
        return super(JSONRemoteModel, cls).parse(response, obj=obj, **kw)


class RestModel(XMLRemoteModel):
    def _name(self):
        return type(self).__name__.lower()

    def _getattr(self, attr):
        return object.__getattribute__(self, attr)

    @classmethod
    def _encode(cls, name):
        name = quote_plus(name).replace('+', '%20')
        return name

    def resource(self, client=None, endpoint=None):
        parent = self._parent
        if parent is None:
            if endpoint is None:
                endpoint = (client or self._client).endpoint
            parent_res = endpoint
        else:
            parent_res = parent.resource(client=client, endpoint=endpoint)
        name = self._name()
        if name is None:
            return parent_res
        return '/'.join([parent_res, self._encode(name)])

    def __eq__(self, other):
        if other is None:
            return False

        if not isinstance(other, type(self)):
            return False

        return self._name() == other._name() and \
            self.parent == other.parent

    def __hash__(self):
        return hash(type(self)) * hash(self._parent) * hash(self._name())

    def _get_schema_name(self):
        if self._schema_name is not utils.notset:
            return self._schema_name

        if isinstance(self._parent, LazyLoad):
            schema = self._parent.get_schema()
        elif isinstance(self._parent, Container):
            schema = self._parent._parent.get_schema()
        else:
            schema = None
        self._schema_name = schema.name if schema is not None else None
        return self._schema_name


class LazyLoad(RestModel):
    __slots__ = '_loaded',

    @cache
    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def __init__(self, **kwargs):
        self._loaded = False
        kwargs.pop('no_cache', None)
        super(LazyLoad, self).__init__(**kwargs)

    def _name(self):
        return self._getattr('name')

    def __getattribute__(self, attr):
        if (
            attr.endswith("_time")
            and attr != "_logview_address_time"
            and type(self).__name__ not in ("Table", "Partition")
            and options.use_legacy_parsedate
        ):
            warnings.warn(
                "We are returning local time instead of UTC time for objects "
                "while the latter is deprecated since PyODPS 0.11.3. Try setting "
                "options.use_legacy_parsedate = False and update your logic.",
                category=DeprecationWarning,
            )
            typ = type(self)
            utils.add_survey_call(".".join([typ.__module__, typ.__name__, attr]) + ":legacy_parsedate")

        val = object.__getattribute__(self, attr)
        if val is None and not self._loaded:
            fields = getattr(type(self), '__fields')
            if attr in fields:
                self.reload()
        return object.__getattribute__(self, attr)

    def reload(self):
        raise NotImplementedError

    def reset(self):
        self._loaded = False

    @property
    def is_loaded(self):
        return self._loaded

    def __repr__(self):
        try:
            r = self._repr()
        except:
            r = None
        if r:
            return r
        else:
            return super(LazyLoad, self).__repr__()

    def _repr(self):
        name = self._name()
        if name:
            return '<%s %s>' % (type(self).__name__, name)
        else:
            raise ValueError

    def __hash__(self):
        return hash((self._name(), self.parent))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return self._name() == other._name() and self.parent == other.parent

    def __getstate__(self):
        return self._name(), self._parent, self._client

    def __setstate__(self, state):
        name, parent, client = state
        self._set_state(name, parent, client)

    def _set_state(self, name, parent, client):
        self.__init__(name=name, _parent=parent, _client=client)

    @property
    def project(self):
        from .project import Project

        cur = self
        while cur is not None and not isinstance(cur, Project):
            cur = cur.parent
        return cur

    def get_schema(self):
        """
        As Table.table_schema already occupied by table_schema result, we need
        an auxiliary method to fix all needs.
        """
        from .schema import Schema

        cur = self
        while cur is not None and not isinstance(cur, Schema):
            cur = cur.parent
        return cur

    @property
    def schema(self):
        return self.get_schema()


class Container(RestModel):
    skip_null = False

    @cache
    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def _get(self, item):
        raise NotImplementedError

    def __getitem__(self, item):
        if isinstance(item, six.string_types):
            item = item.strip()
            if not item:
                raise ValueError("Empty string not supported")
            return self._get(item)
        raise ValueError('Unsupported getitem value: %s' % item)

    @del_cache
    def __delitem__(self, key):
        pass

    def __contains__(self, item):
        raise NotImplementedError

    def __getstate__(self):
        return self._parent, self._client

    def __setstate__(self, state):
        parent, client = state
        self.__init__(_parent=parent, _client=client)


class Iterable(Container):
    __slots__ = '_iter',

    def __init__(self, **kwargs):
        super(Iterable, self).__init__(**kwargs)
        self._iter = iter(self)

    def __iter__(self):
        raise NotImplementedError

    def __next__(self):
        return next(self._iter)

    next = __next__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import types as tps
from decimal import Decimal

from .compiler import SQLAlchemyCompiler
from . import analyzer as ana
from . import rewriter as rwr
from .types import df_schema_to_sqlalchemy_columns
from ..core import ExecuteNode, Engine
from ..errors import CompileError
from ..utils import write_table, refresh_dynamic
from ..frame import ResultFrame
from ..context import context
from ... import DataFrame
from ...utils import is_source_collection, is_constant_scalar
from ...expr.expressions import *
from ...expr.core import ExprDAG
from ...expr.dynamic import DynamicMixin
from ...types import DynamicSchema, Unknown
from ...backends.odpssql.types import df_schema_to_odps_schema, df_type_to_odps_type
from ....types import PartitionSpec
from ....utils import gen_temp_table
from ....models import TableSchema, Partition

logger = logging.getLogger(__name__)


class SQLExecuteNode(ExecuteNode):
    def _sql(self):
        raise NotImplementedError

    def __repr__(self):
        buf = six.StringIO()

        sql = self._sql()
        buf.write('MPP SQL compiled: \n\n')
        buf.write(sql)

        return buf.getvalue()

    def _repr_html_(self):
        buf = six.StringIO()

        sql = self._sql()
        buf.write('<h4>MPP SQL compiled</h4>')
        buf.write('<code>%s</code>' % sql)

        return buf.getvalue()


_engine_to_connections = {}


class SQLAlchemyEngine(Engine):
    def __init__(self, odps=None):
        self._odps = odps

    def _new_execute_node(self, expr_dag):
        node = SQLExecuteNode(expr_dag)

        def _sql(*_):
            return self._compile_sql(expr_dag)

        def verify(*_):
            try:
                _sql(*_)
                return True
            except NotImplementedError:
                return False

        node._sql = tps.MethodType(_sql, node)
        node.verify = tps.MethodType(verify, node)
        return node

    def _new_analyzer(self, expr_dag, on_sub=None):
        return ana.Analyzer(expr_dag, on_sub=on_sub)

    def _new_rewriter(self, expr_dag):
        return rwr.Rewriter(expr_dag)

    def _compile_sql(self, expr_dag):
        self._rewrite(expr_dag)

        sa = self._compile(expr_dag.root, expr_dag)
        return self._sa_to_sql(sa)

    def _sa_to_sql(self, sa):
        try:
            return sa.compile(compile_kwargs={"literal_binds": True})
        except NotImplementedError:
            return sa.compile()

    def _status_ui(self, ui):
        ui.status('Try to execute by sqlalchemy...', clear_keys=True)

    @classmethod
    def _get_or_create_conn(cls, engine):
        if engine in _engine_to_connections:
            return _engine_to_connections[engine]
        conn = engine.connect()
        _engine_to_connections[engine] = conn
        return conn

    def _get_cast_func(self, tp):
        df_types_to_builtin_types = {
            types.Integer: int,
            types.Float: float
        }
        for df_type, builtin_type in six.iteritems(df_types_to_builtin_types):
            if isinstance(tp, df_type):
                return lambda value: builtin_type(value)
        return lambda value: value

    def _get_sa_table(self, table_name, engine, schema):
        import sqlalchemy

        metadata = sqlalchemy.MetaData(bind=engine)
        columns = df_schema_to_sqlalchemy_columns(schema, engine=engine)
        table = sqlalchemy.Table(table_name, metadata, *columns, extend_existing=True)

        return table

    def _get_result(self, table_name, engine, schema, head, tail):
        import sqlalchemy

        table = self._get_sa_table(table_name, engine, schema)
        sa = sqlalchemy.select([table.alias('t1')])
        if head:
            sa = sa.limit(head)
        elif tail:
            count = sa.alias('t2').count().execute().scalar()
            skip = max(count - tail, 0)
            if skip:
                sa = sa.offset(skip)

        conn = self._get_or_create_conn(engine)
        return conn.execute(sa)

    @classmethod
    def _create_table(cls, table_name, sa, expr):
        from .ext import SACreateTempTableAs

        return SACreateTempTableAs(table_name, sa)

    def _convert_result(self, result, schema):
        res = [list(r) for r in result]
        if len(res) > 0:
            record = res[0]
            for i, r in enumerate(record):
                if schema[i].type != types.decimal and isinstance(r, Decimal):
                    cast = self._get_cast_func(schema[i].type)
                    for r in res:
                        r[i] = cast(r[i])

        return res

    def _run(self, sa, ui, expr_dag, src_expr, progress_proportion=1,
             head=None, tail=None, fetch=True, tmp_table_name=None,
             execution_options=None):
        self._status_ui(ui)

        schema = expr_dag.root.schema
        execution_options = dict() if execution_options is None else execution_options
        try:
            if isinstance(src_expr, (CollectionExpr, SequenceExpr)):
                if tmp_table_name is None:
                    tmp_table_name = gen_temp_table()
                to_execute = self._create_table(tmp_table_name, sa, expr_dag.root)
                logger.info('Sql compiled:\n' + self._sa_to_sql(to_execute))
                conn = self._get_or_create_conn(sa.bind)
                conn.execution_options(**execution_options).execute(to_execute)

                if fetch:
                    result = self._get_result(tmp_table_name, sa.bind, schema, head, tail)
                    res = self._convert_result(result, schema)
                    return tmp_table_name, res
                else:
                    return tmp_table_name, None
            else:
                logger.info('Sql compiled:\n' + self._sa_to_sql(sa))

                conn = self._get_or_create_conn(sa.bind)
                res = conn.execution_options(**execution_options).execute(sa).scalar()
                if src_expr.dtype != types.decimal and isinstance(res, Decimal):
                    return self._get_cast_func(src_expr.dtype)(res)
                return res
        finally:
            ui.inc(progress_proportion)

    def _cache(self, expr_dag, dag, expr, **kwargs):
        if is_source_collection(expr_dag.root) or \
                is_constant_scalar(expr_dag.root):
            return

        execute_dag = ExprDAG(expr_dag.root, dag=expr_dag)

        if isinstance(expr, CollectionExpr):
            table_name = gen_temp_table()
            table = self._get_table(table_name, expr_dag)
            root = expr_dag.root
            sub = CollectionExpr(_source_data=table, _schema=expr.schema)
            sub.add_deps(root)
            expr_dag.substitute(root, sub)

            kw = dict(kwargs)

            execute_node = self._execute(execute_dag, dag, expr,
                                         execute_kw={'fetch': False, 'temp_table_name': table_name},
                                         **kw)

            def callback(res):
                if isinstance(expr, DynamicMixin):
                    sub._schema = res.schema
                    refresh_dynamic(sub, expr_dag)

            execute_node.callback = callback
        else:
            assert isinstance(expr, Scalar)  # sequence is not cache-able

            class ValueHolder(object):
                pass

            sub = Scalar(_value_type=expr.dtype)
            sub._value = ValueHolder()
            root = expr_dag.root
            sub.add_deps(root)
            expr_dag.substitute(root, sub)

            execute_node = self._execute(execute_dag, dag, expr, **kwargs)

            def callback(res):
                sub._value = res

            execute_node.callback = callback

        return sub, execute_node

    def _compile(self, expr, dag):
        compiler = SQLAlchemyCompiler(dag)
        return compiler.compile(expr)

    def _get_table(self, table_name, expr_dag, bind=None):
        if bind is None:
            bind = next(e for e in expr_dag.traverse()
                        if is_source_collection(e) and e._source_data.bind)._source_data.bind

        return self._get_sa_table(table_name, bind, expr_dag.root.schema)

    def _do_execute(self, expr_dag, expr, ui=None, progress_proportion=1,
                    head=None, tail=None, **kwargs):
        expr_dag = self._convert_table(expr_dag)
        self._rewrite(expr_dag)

        src_expr = expr
        expr = expr_dag.root

        if isinstance(expr, Scalar) and expr.value is not None:
            ui.inc(progress_proportion)
            return expr.value

        sqlalchemy_expr = self._compile(expr, expr_dag)

        fetch = kwargs.pop('fetch', True)
        temp_table_name = kwargs.pop('temp_table_name', None)
        execution_options = kwargs.pop('execution_options',
                                       options.df.sqlalchemy.execution_options)
        result = self._run(sqlalchemy_expr, ui, expr_dag, src_expr,
                           progress_proportion=progress_proportion,
                           head=head, tail=tail, fetch=fetch,
                           tmp_table_name=temp_table_name,
                           execution_options=execution_options)

        if not isinstance(src_expr, Scalar):
            # reset schema
            if isinstance(src_expr, CollectionExpr) and \
                    (isinstance(src_expr._schema, DynamicSchema) or
                         any(isinstance(col.type, Unknown) for col in src_expr._schema.columns)):
                src_expr._schema = expr_dag.root.schema
            table_name, result = result
            table = self._get_table(table_name, expr_dag, sqlalchemy_expr.bind)
            context.cache(src_expr, table)
            if fetch:
                return ResultFrame(result, schema=expr_dag.root.schema)
            else:
                return table
        else:
            context.cache(src_expr, result)
            return result

    def _do_persist(self, expr_dag, expr, name, partitions=None, partition=None,
                    odps=None, project=None, ui=None,
                    progress_proportion=1, execute_percent=0.5, lifecycle=None,
                    overwrite=True, drop_table=False, create_table=True,
                    drop_partition=False, create_partition=False, cast=False, **kwargs):
        expr_dag = self._convert_table(expr_dag)
        self._rewrite(expr_dag)

        src_expr = expr
        expr = expr_dag.root
        odps = odps or self._odps

        try:
            import pandas
        except ImportError:
            raise DependencyNotInstalledError('persist requires for pandas')

        df = self._do_execute(expr_dag, src_expr, ui=ui,
                              progress_proportion=progress_proportion * execute_percent, **kwargs)
        schema = TableSchema(columns=df.columns)

        if partitions is not None:
            if drop_partition:
                raise ValueError('Cannot drop partitions when specify `partitions`')
            if create_partition:
                raise ValueError('Cannot create partitions when specify `partitions`')

            if isinstance(partitions, tuple):
                partitions = list(partitions)
            if not isinstance(partitions, list):
                partitions = [partitions, ]

            for p in partitions:
                if p not in schema:
                    raise ValueError(
                            'Partition field(%s) does not exist in DataFrame schema' % p)

            schema = df_schema_to_odps_schema(schema)
            columns = [c for c in schema.columns if c.name not in partitions]
            ps = [Partition(name=t, type=schema.get_type(t)) for t in partitions]
            schema = TableSchema(columns=columns, partitions=ps)

            if odps.exist_table(name, project=project) or not create_table:
                t = odps.get_table(name, project=project)
            else:
                t = odps.create_table(name, schema, project=project)
        elif partition is not None:
            if odps.exist_table(name, project=project) or not create_table:
                t = odps.get_table(name, project=project)
                partition = self._get_partition(partition, t)

                if drop_partition:
                    t.delete_partition(partition, if_exists=True)
                if create_partition:
                    t.create_partition(partition, if_not_exists=True)
            else:
                partition = self._get_partition(partition)
                project_obj = odps.get_project(project)
                column_names = [n for n in expr.schema.names if n not in partition]
                column_types = [
                    df_type_to_odps_type(expr.schema[n].type, project=project_obj)
                    for n in column_names
                ]
                partition_names = [n for n in partition.keys]
                partition_types = ['string'] * len(partition_names)
                t = odps.create_table(
                    name, TableSchema.from_lists(
                        column_names, column_types, partition_names, partition_types
                    ),
                    project=project,
                )
                if create_partition is None or create_partition is True:
                    t.create_partition(partition)
        else:
            if drop_partition:
                raise ValueError('Cannot drop partition for non-partition table')
            if create_partition:
                raise ValueError('Cannot create partition for non-partition table')

            if odps.exist_table(name, project=project) or not create_table:
                t = odps.get_table(name, project=project)
                if t.table_schema.partitions:
                    raise CompileError('Cannot insert into partition table %s without specifying '
                                       '`partition` or `partitions`.')
            else:
                t = odps.create_table(name, df_schema_to_odps_schema(schema), project=project)

        write_table(df, t, ui=ui, cast=cast, overwrite=overwrite, partitions=partitions, partition=partition,
                    progress_proportion=progress_proportion * (1 - execute_percent))

        if partition:
            filters = []
            df = DataFrame(t)
            for k in partition.keys:
                filters.append(df[k] == partition[k])
            return df.filter(*filters)
        return DataFrame(t)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sqlalchemy.sql.expression import FunctionElement, Executable, ClauseElement
from sqlalchemy.ext.compiler import compiles


class SALog(FunctionElement):
    def __init__(self, name, base, *args, **kwargs):
        self.name = name
        self.base = base
        super(SALog, self).__init__(*args, **kwargs)


@compiles(SALog)
def visit_log(log, compiler, **_):
    if log.base is None:
        return 'ln(%s)' % compiler.process(log.clauses)
    else:
        return 'log(%s)' % compiler.process(log.clauses)


@compiles(SALog, 'mssql')
def visit_log(log, compiler, **_):
    if log.base is None:
        return 'log(%s)' % compiler.process(log.clauses)
    else:
        return 'log(%s)' % compiler.process(
            log.clauses._constructor(log.clauses.clauses[::-1]))  # in mssql, log(num, base)


class SATruncate(FunctionElement):
    def __init__(self, name, *args, **kwargs):
        self.name = name
        super(SATruncate, self).__init__(*args, **kwargs)


@compiles(SATruncate, 'postgresql')
@compiles(SATruncate, 'oracle')
def visit_truncate(trunc, compiler, **_):
    return 'trunc(%s)' % compiler.process(trunc.clauses)


@compiles(SATruncate, 'mysql')
def visit_truncate(trunc, compiler, **_):
    return 'truncate(%s)' % compiler.process(trunc.clauses)


@compiles(SATruncate, 'mssql')
def visit_truncate(trunc, compiler, **_):
    return 'round(%s, 1)' % compiler.process(trunc.clauses)


class SACreateTempTableAs(Executable, ClauseElement):

    def __init__(self, name, query):
        self.name = name
        self.query = query
        if hasattr(self.query, 'bind') and self.query.bind is not None \
                and self.bind is None:
            self._bind = query.bind


@compiles(SACreateTempTableAs, 'mysql')
def create_temp_table_as(element, compiler, **kw):
    return 'CREATE TEMPORARY TABLE %s AS %s' % (
        element.name,
        compiler.process(element.query),
    )


@compiles(SACreateTempTableAs)
def create_temp_table_as(element, compiler, **kw):
    return 'CREATE TEMP TABLE %s AS %s' % (
        element.name,
        compiler.process(element.query),
    )



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import odps.counters

_origin_int = int
_annotated_classes = {}

__all__ = ['get_execution_context', 'ExecutionContext', 'BaseUDAF', 'BaseUDTF', 'int', 'annotate', 'get_annotation']

class ExecutionContext(object):
    counters = odps.counters.Counters()

    def get_counters(self):
        return self.counters

    def get_counters_as_json_string(self):
        return self.counters.to_json_string()

class BaseUDAF(object):
    pass

class BaseUDTF(object):
    def forward(self, *args):
        pass

    def close(self):
        pass

def get_execution_context():
    return ExecutionContext()

def int(v, silent=True):
    v = _int(v)
    if type(v) is long:
        if silent:
            return None
        else:
            raise OverflowError('Python int too large to convert to bigint: %s' % v)
    return v

def annotate(prototype):
    def ann(clz):
        _annotated_classes[clz] = prototype
        return clz
    return ann

def get_annotation(clz):
    return _annotated_classes.get(clz)



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
import warnings

from odps.compat import six
from odps.utils import with_wait_argument

from .config import options
from .errors import CupidMasterTimeoutError
from .rpc import CupidTaskServiceRpcChannel
from .utils import get_environ

try:
    from .proto import cupidtaskparam_pb2 as task_param_pb
except TypeError:
    warnings.warn('Cannot import protos from pycupid: '
        'consider upgrading your protobuf python package.', ImportWarning)
    raise ImportError

logger = logging.getLogger(__name__)

_CUPID_CONF_PREFIXES = 'odps cupid'.split()


class CupidSession(object):
    def __init__(self, odps=None, project=None):
        from .runtime import context
        from odps import ODPS

        self._context = context()

        if odps is not None:
            self.odps = odps
        else:
            self.odps = ODPS.from_global()
        self.project = project if project is not None else self.odps.project

        self.lookup_name = get_environ('META_LOOKUP_NAME', '')
        self.running = False
        self.save_id = None
        self.job_running_event = threading.Event()

        self._kube_url = None

    def check_running(self, timeout=None):
        timeout = timeout or options.cupid.wait_am_start_time
        if not self.running:
            self.job_running_event.wait(timeout)
        return self.running

    @property
    def account(self):
        # todo add support for bearer token
        return self.odps.account

    def job_conf(self, conf=None):
        conf_items = {
            'odps.task.major.version': options.cupid.major_task_version,
            'odps.access.id': getattr(self.odps.account, 'access_id', None),
            'odps.access.key': getattr(self.odps.account, 'secret_access_key', None),
            'odps.end.point': self.odps.endpoint,
            'odps.project.name': self.project,
            'odps.moye.am.cores': '400',
            'odps.cupid.proxy.end.point': options.cupid.proxy_endpoint,
        }

        if conf:
            conf_items.update(conf)
        for k, v in six.iteritems(options.cupid.settings or {}):
            if any(k.startswith(pf) for pf in _CUPID_CONF_PREFIXES):
                conf_items[k] = v
        conf_obj_items = [task_param_pb.JobConfItem(key=k, value=str(v))
                          for k, v in six.iteritems(conf_items) if v is not None]
        return task_param_pb.JobConf(jobconfitem=conf_obj_items)

    def get_proxy_token(self, instance, app_name, expired_in_hours):
        if hasattr(instance, 'id'):
            instance = instance.id

        task_operator = task_param_pb.CupidTaskOperator(moperator='GetCupidProxyToken')
        proxy_token_request = task_param_pb.CupidProxyTokenRequest(
            instanceId=instance,
            appName=app_name,
            expiredInHours=expired_in_hours,
        )
        task_param = task_param_pb.CupidTaskParam(
            mcupidtaskoperator=task_operator,
            cupidProxyTokenRequest=proxy_token_request,
        )
        channel = CupidTaskServiceRpcChannel(self)
        inst = channel.submit_job(task_param, 'eAsyncNotFuxiJob', session=self)
        return channel.wait_cupid_instance(inst)

    def get_proxied_url(self, instance, app_name, expired_in_hours=None):
        expired_in_hours = expired_in_hours or 30 * 24
        return 'http://%s.%s' % (self.get_proxy_token(instance, app_name, expired_in_hours),
                                 options.cupid.proxy_endpoint)

    @with_wait_argument
    def start_kubernetes(self, async_=False, priority=None, running_cluster=None,
                         proxy_endpoint=None, major_task_version=None,
                         app_command=None, app_image=None, resources=None, **kw):
        priority = priority if priority is not None else options.priority
        if priority is None and options.get_priority is not None:
            priority = options.get_priority(self.odps)
        menginetype = options.cupid.engine_running_type

        if proxy_endpoint is not None:
            options.cupid.proxy_endpoint = proxy_endpoint
        if major_task_version is not None:
            options.cupid.major_task_version = major_task_version

        runtime_endpoint = kw.pop('runtime_endpoint', None)
        task_operator = task_param_pb.CupidTaskOperator(moperator='startam', menginetype=menginetype)
        task_name = kw.pop('task_name', None)

        if len(kw) > 0:
            raise ValueError('Got unexpected arguments: {}'.format(list(kw.keys())[0]))

        kub_conf = {
            'odps.cupid.kube.master.mode': options.cupid.kube.master_mode,
            'odps.cupid.master.type': options.cupid.master_type,
            'odps.cupid.engine.running.type': menginetype,
            'odps.cupid.job.capability.duration.hours': options.cupid.job_duration_hours,
            'odps.cupid.channel.init.timeout.seconds': options.cupid.channel_init_timeout_seconds,
            'odps.moye.runtime.type': options.cupid.application_type,
            'odps.runtime.end.point': runtime_endpoint or options.cupid.runtime.endpoint,
            'odps.cupid.resources': ','.join(resources or []),
            'odps.cupid.kube.appmaster.cmd': app_command,
            'odps.cupid.kube.appmaster.image': app_image
        }
        if running_cluster:
            kub_conf['odps.cupid.task.running.cluster'] = running_cluster
        if options.cupid.container_node_label is not None:
            kub_conf['odps.cupid.container.node.label'] = options.cupid.container_node_label
        if options.cupid.master.virtual_resource is not None:
            kub_conf['odps.cupid.master.virtual.resource'] = options.cupid.master.virtual_resource

        task_param = task_param_pb.CupidTaskParam(
            jobconf=self.job_conf(conf=kub_conf),
            mcupidtaskoperator=task_operator,
        )

        retrial = 0

        while True:
            retrial += 1
            channel = CupidTaskServiceRpcChannel(self)
            inst = channel.submit_job(task_param, 'eHasFuxiJob', with_resource=True, priority=priority,
                                    running_cluster=running_cluster, session=self, task_name=task_name)
            if async_:
                return inst
            else:
                try:
                    return self.get_instance_kube(inst)
                except CupidMasterTimeoutError:
                    if retrial == 3:
                        raise

    def get_instance_kube_api(self, inst, expired_in_hours=None):
        if self._kube_url is None:
            CupidTaskServiceRpcChannel.wait_cupid_instance(inst, 'running')
            self._kube_url = self.get_proxied_url(inst, '', expired_in_hours or 30 * 24)
        return self._kube_url

    def get_instance_kube(self, inst, expired_in_hours=None):
        from kubernetes import client
        config = client.Configuration()
        config.host = self.get_instance_kube_api(inst, expired_in_hours)
        return client.ApiClient(config)


from .io import *

__all__ = ['CupidSession', 'CupidTableUploadSession', 'CupidTableDownloadSession']



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libcpp.vector cimport vector

ctypedef object (*_VALIDATE_FUNC)(object val, int64_t max_field_size)

cdef class SchemaSnapshot:
    cdef list _columns
    cdef int _col_count
    cdef int _partition_col_count
    cdef list _col_types
    cdef vector[int] _col_type_ids
    cdef vector[int] _col_is_partition
    cdef vector[int] _col_nullable
    cdef vector[_VALIDATE_FUNC] _col_validators

    cdef object validate_value(self, int i, object val, int64_t max_field_size)

cdef class BaseRecord:
    cdef list _c_columns, _c_values
    cdef dict _c_name_indexes
    cdef SchemaSnapshot _c_schema_snapshot
    cdef int64_t _max_field_size

    cpdef object get_by_name(self, object name)
    cpdef set_by_name(self, object name, object value)

    cdef size_t _get_non_partition_col_count(self)

    cpdef object _get(self, int i)
    cpdef _set(self, int i, object value)
    cpdef _sets(self, object values)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import functools
import json
import struct
import warnings
from collections import OrderedDict
from decimal import Decimal
from io import IOBase, BytesIO, StringIO

from ...types import PartitionSpec

try:
    import numpy as np
except ImportError:
    np = None
try:
    import pandas as pd
except ImportError:
    pd = None
try:
    import pyarrow as pa
except (AttributeError, ImportError):
    pa = None


from ... import utils, types, compat
from ...errors import DatetimeOverflowError
from ...models import Record
from ...readers import AbstractRecordReader
from ...config import options
from ..pb.decoder import Decoder
from ..pb.errors import DecodeError
from ..checksum import Checksum
from ..wireconstants import ProtoWireConstants
from .types import odps_schema_to_arrow_schema

try:
    if not options.force_py:
        from .reader_c import TunnelRecordReader, convert_legacy_decimal_bytes
    else:
        TunnelRecordReader = convert_legacy_decimal_bytes = None
except ImportError as e:
    if options.force_c:
        raise e
    TunnelRecordReader = convert_legacy_decimal_bytes = None

if TunnelRecordReader is None:
    class TunnelRecordReader(AbstractRecordReader):

        def __init__(self, schema, input_stream, columns=None, partition_spec=None):
            self._schema = schema
            if columns is None:
                self._columns = self._schema.columns
            else:
                self._columns = [self._schema[c] for c in columns]
            self._reader = Decoder(input_stream)
            self._crc = Checksum()
            self._crccrc = Checksum()
            self._curr_cursor = 0
            self._read_limit = options.table_read_limit
            self._to_datetime = utils.MillisecondsConverter().from_milliseconds
            self._to_datetime_utc = utils.MillisecondsConverter(
                local_tz=False
            ).from_milliseconds
            self._to_date = utils.to_date
            self._partition_spec = PartitionSpec(partition_spec) if partition_spec else None

        def _mode(self):
            return 'py'

        @property
        def count(self):
            return self._curr_cursor

        @property
        def schema(self):
            return self._schema

        def _read_field(self, data_type):
            if data_type == types.float_:
                val = self._reader.read_float()
                self._crc.update_float(val)
            elif data_type == types.double:
                val = self._reader.read_double()
                self._crc.update_double(val)
            elif data_type == types.boolean:
                val = self._reader.read_bool()
                self._crc.update_bool(val)
            elif data_type in types.integer_types:
                val = self._reader.read_sint64()
                self._crc.update_long(val)
            elif data_type == types.string:
                val = self._reader.read_string()
                self._crc.update(val)
            elif data_type == types.binary:
                val = self._reader.read_string()
                self._crc.update(val)
            elif data_type == types.datetime:
                val = self._reader.read_sint64()
                self._crc.update_long(val)
                try:
                    val = self._to_datetime(val)
                except DatetimeOverflowError:
                    if not options.tunnel.overflow_date_as_none:
                        raise
                    val = None
            elif data_type == types.date:
                val = self._reader.read_sint64()
                self._crc.update_long(val)
                val = self._to_date(val)
            elif data_type == types.timestamp or data_type == types.timestamp_ntz:
                to_datetime = self._to_datetime_utc if data_type == types.timestamp_ntz else self._to_datetime
                l_val = self._reader.read_sint64()
                self._crc.update_long(l_val)
                nano_secs = self._reader.read_sint32()
                self._crc.update_int(nano_secs)
                if pd is None:
                    raise ImportError('To use TIMESTAMP in pyodps, you need to install pandas.')
                try:
                    val = pd.Timestamp(to_datetime(l_val * 1000)) + pd.Timedelta(nanoseconds=nano_secs)
                except DatetimeOverflowError:
                    if not options.tunnel.overflow_date_as_none:
                        raise
                    val = None
            elif data_type == types.interval_day_time:
                l_val = self._reader.read_sint64()
                self._crc.update_long(l_val)
                nano_secs = self._reader.read_sint32()
                self._crc.update_int(nano_secs)
                if pd is None:
                    raise ImportError('To use INTERVAL_DAY_TIME in pyodps, you need to install pandas.')
                val = pd.Timedelta(seconds=l_val, nanoseconds=nano_secs)
            elif data_type == types.interval_year_month:
                l_val = self._reader.read_sint64()
                self._crc.update_long(l_val)
                return compat.Monthdelta(l_val)
            elif data_type == types.json:
                sval = self._reader.read_string()
                val = json.loads(sval)
                self._crc.update(sval)
            elif isinstance(data_type, (types.Char, types.Varchar)):
                val = self._reader.read_string()
                self._crc.update(val)
            elif isinstance(data_type, types.Decimal):
                val = self._reader.read_string()
                self._crc.update(val)
            elif isinstance(data_type, types.Array):
                val = self._read_array(data_type.value_type)
            elif isinstance(data_type, types.Map):
                keys = self._read_array(data_type.key_type)
                values = self._read_array(data_type.value_type)
                val = collections.OrderedDict(zip(keys, values))
            elif isinstance(data_type, types.Struct):
                val = self._read_struct(data_type)
            else:
                raise IOError('Unsupported type %s' % data_type)
            return val

        def _read_array(self, value_type):
            res = []

            size = self._reader.read_uint32()
            for _ in range(size):
                if self._reader.read_bool():
                    res.append(None)
                else:
                    res.append(self._read_field(value_type))

            return res

        def _read_struct(self, value_type):
            res_list = [None] * len(value_type.field_types)
            for idx, field_type in enumerate(value_type.field_types.values()):
                if not self._reader.read_bool():
                    res_list[idx] = self._read_field(field_type)
            if options.struct_as_dict:
                return OrderedDict(zip(value_type.field_types.keys(), res_list))
            else:
                return value_type.namedtuple_type(*res_list)

        def read(self):
            if self._read_limit is not None and self.count >= self._read_limit:
                warnings.warn(
                    'Number of lines read via tunnel already reaches the limitation.',
                    RuntimeWarning,
                )
                return None

            record = Record(self._columns, max_field_size=(1 << 63) - 1)

            while True:
                index, _ = self._reader.read_field_number_and_wire_type()

                if index == 0:
                    continue
                if index == ProtoWireConstants.TUNNEL_END_RECORD:
                    checksum = utils.long_to_int(self._crc.getvalue())
                    if int(self._reader.read_uint32()) != utils.int_to_uint(checksum):
                        raise IOError('Checksum invalid')
                    self._crc.reset()
                    self._crccrc.update_int(checksum)
                    break

                if index == ProtoWireConstants.TUNNEL_META_COUNT:
                    if self.count != self._reader.read_sint64():
                        raise IOError('count does not match')
                    idx_of_checksum, _ = self._reader.read_field_number_and_wire_type()
                    if ProtoWireConstants.TUNNEL_META_CHECKSUM != idx_of_checksum:
                        raise IOError('Invalid stream data.')
                    if int(self._crccrc.getvalue()) != self._reader.read_uint32():
                        raise IOError('Checksum invalid.')

                    return

                if index > len(self._columns):
                    raise IOError('Invalid protobuf tag. Perhaps the datastream '
                                  'from server is crushed.')

                self._crc.update_int(index)

                i = index - 1
                record[i] = self._read_field(self._columns[i].type)

            if self._partition_spec is not None:
                for k, v in self._partition_spec.items():
                    try:
                        record[k] = v
                    except KeyError:
                        # skip non-existing fields
                        pass

            self._curr_cursor += 1
            return record

        def __next__(self):
            record = self.read()
            if record is None:
                raise StopIteration
            return record

        next = __next__

        def reads(self):
            return self.__iter__()

        @property
        def n_bytes(self):
            return len(self._reader)

        def get_total_bytes(self):
            return self.n_bytes

        def close(self):
            if hasattr(self._schema, 'close'):
                self._schema.close()

        def __enter__(self):
            return self

        def __exit__(self, *_):
            self.close()


class ArrowStreamReader(IOBase):
    def __init__(self, raw_reader, arrow_schema):
        self._reader = raw_reader

        self._crc = Checksum()
        self._crccrc = Checksum()
        self._pos = 0
        self._chunk_size = None

        self._buffers = collections.deque()
        self._buffers.append(
            BytesIO(arrow_schema.serialize().to_pybytes())
        )

    def readable(self):
        return True

    @staticmethod
    def _read_unint32(b):
        return struct.unpack('!I', b)

    def _read_chunk_size(self):
        try:
            i = self._read_unint32(self._reader.read(4))
            self._pos += 4
            return i[0]  # unpack() result is a 1-element tuple.
        except struct.error as e:
            raise DecodeError(e)

    def _read_chunk(self):
        read_size = self._chunk_size + 4
        b = self._reader.read(read_size)
        if 0 < len(b) < 4:
            raise IOError('Checksum invalid')
        self._pos += len(b)
        self._crc.update(b[:-4])
        self._crccrc.update(b[:-4])
        return b

    def _fill_next_buffer(self):
        if self._chunk_size is None:
            self._chunk_size = self._read_chunk_size()

        b = self._read_chunk()
        data = b[:-4]
        crc_data = b[-4:]
        if len(b) == 0:
            return

        if len(b) < self._chunk_size + 4:
            # is last chunk
            read_checksum = self._read_unint32(crc_data)[0]
            checksum = int(self._crccrc.getvalue())
            if checksum != read_checksum:
                raise IOError('Checksum invalid')
            self._pos += len(data) + 4
            self._buffers.append(BytesIO(data))
            self._crccrc.reset()
        else:
            checksum = int(self._crc.getvalue())
            read_checksum = self._read_unint32(crc_data)[0]
            if checksum != read_checksum:
                raise IOError('Checksum invalid')
            self._crc.reset()
            self._buffers.append(BytesIO(data))

    def read(self, nbytes=None):
        tot_size = 0
        bufs = []
        while nbytes is None or tot_size < nbytes:
            if not self._buffers:
                self._fill_next_buffer()
                if not self._buffers:
                    break

            to_read = nbytes - tot_size if nbytes is not None else None
            buf = self._buffers[0].read(to_read)
            if not buf:
                self._buffers.popleft()
            else:
                bufs.append(buf)
                tot_size += len(buf)
        if len(bufs) == 1:
            return bufs[0]
        return b''.join(bufs)

    def close(self):
        if hasattr(self._reader, 'close'):
            self._reader.close()


class TunnelArrowReader(object):
    def __init__(
        self, schema, input_stream, columns=None, use_ipc_stream=False
    ):
        if pa is None:
            raise ValueError("To use arrow reader you need to install pyarrow")

        self._schema = schema
        self._columns = columns

        arrow_schema = odps_schema_to_arrow_schema(schema)
        if columns is None:
            self._arrow_schema = arrow_schema
        else:
            self._arrow_schema = pa.schema([s for s in arrow_schema if s.name in columns])

        if use_ipc_stream:
            self._reader = input_stream
        else:
            self._reader = ArrowStreamReader(input_stream, self._arrow_schema)

        self._pos = 0
        self._arrow_stream = None
        self._to_datetime = utils.MillisecondsConverter().from_milliseconds
        self._read_limit = options.table_read_limit

        self.closed = False

        self._pd_column_converters = dict()
        for col in schema.simple_columns:
            if isinstance(col.type, (types.Map, types.Array, types.Struct)):
                self._pd_column_converters[col.name] = ArrowRecordFieldConverter(
                    col.type, convert_ts=False
                )

    @property
    def schema(self):
        return self._schema

    def read_next_batch(self):
        if self._reader is None:
            return None

        if self._arrow_stream is None:
            self._arrow_stream = pa.ipc.open_stream(self._reader)

        if self._read_limit is not None and self._pos >= self._read_limit:
            warnings.warn(
                'Number of lines read via tunnel already reaches the limitation.',
                RuntimeWarning,
            )
            return None

        try:
            batch = self._arrow_stream.read_next_batch()
            self._pos += batch.num_rows
        except StopIteration:
            return None

        col_names = self._columns or batch.schema.names
        col_to_array = dict()
        col_name_set = set(col_names)

        for name, arr in zip(batch.schema.names, batch.columns):
            if name not in col_name_set:
                continue
            if arr.type == pa.timestamp('ms'):
                col_to_array[name] = np.vectorize(self._to_datetime)(
                    arr.cast('int64').to_numpy()
                )
            else:
                col_to_array[name] = arr
        arrays = [col_to_array[name] for name in col_names]
        return pa.RecordBatch.from_arrays(arrays, names=col_names)

    def read(self):
        batches = []
        while True:
            batch = self.read_next_batch()
            if batch is None:
                break
            batches.append(batch)

        if not batches:
            return self._arrow_schema.empty_table()
        return pa.Table.from_batches(batches)

    def __iter__(self):
        return self

    def __next__(self):
        batch = self.read_next_batch()
        if batch is None:
            raise StopIteration
        return batch

    @property
    def n_bytes(self):
        return self._pos

    def get_total_bytes(self):
        return self.n_bytes

    def close(self):
        if hasattr(self._reader, 'close'):
            self._reader.close()

    def _convert_batch_to_pandas(self, batch):
        series_list = []
        if not self._pd_column_converters:
            return batch.to_pandas()
        for col_name, arrow_column in zip(batch.schema.names, batch.columns):
            if col_name not in self._pd_column_converters:
                series_list.append(arrow_column.to_pandas())
            else:
                try:
                    series = arrow_column.to_pandas()
                except pa.ArrowNotImplementedError:
                    series = pd.Series(arrow_column.to_pylist(), name=col_name)
                series_list.append(series.map(self._pd_column_converters[col_name]))
        return pd.concat(series_list, axis=1)

    def to_pandas(self):
        import pandas as pd

        batches = []
        while True:
            batch = self.read_next_batch()
            if batch is None:
                break
            batches.append(self._convert_batch_to_pandas(batch))

        if not batches:
            return self._arrow_schema.empty_table().to_pandas()
        return pd.concat(batches, axis=0)

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()


_reflective = lambda x: x


if convert_legacy_decimal_bytes is None:
    def convert_legacy_decimal_bytes(value):
        """
        Legacy decimal memory layout:
            int8_t  mNull;
            int8_t  mSign;
            int8_t  mIntg;
            int8_t  mFrac; only 0, 1, 2
            int32_t mData[6];
            int8_t mPadding[4]; //For Memory Align
        """
        if value is None:
            return None

        is_null, sign, intg, frac = struct.unpack("<4b", value[:4])
        if is_null:  # pragma: no cover
            return None
        if intg + frac == 0:
            return Decimal("0")

        sio = BytesIO() if compat.PY27 else StringIO()
        if sign > 0:
            sio.write("-")
        intg_nums = struct.unpack("<%dI" % intg, value[12: 12 + intg * 4])
        intg_val = "".join("%09d" % d for d in reversed(intg_nums)).lstrip("0")
        sio.write(intg_val or "0")
        if frac > 0:
            sio.write(".")
            frac_nums = struct.unpack("<%dI" % frac, value[12 - frac * 4: 12])
            sio.write("".join("%09d" % d for d in reversed(frac_nums)))
        return Decimal(sio.getvalue())


class ArrowRecordFieldConverter(object):
    _sensitive_types = (
        types.Datetime,
        types.Timestamp,
        types.TimestampNTZ,
        types.Array,
        types.Map,
        types.Struct,
    )

    def __init__(self, odps_type, arrow_type=None, convert_ts=True):
        self._mills_converter = utils.MillisecondsConverter()
        self._mills_converter_utc = utils.MillisecondsConverter(local_tz=False)
        self._convert_ts = convert_ts

        self._converter = self._build_converter(odps_type, arrow_type)

    def _convert_datetime(self, value):
        if value is None:
            return None
        mills = self._mills_converter.to_milliseconds(value)
        return self._mills_converter.from_milliseconds(mills)

    def _convert_ts_timestamp(self, value, ntz=False):
        if value is None:
            return None

        if not ntz:
            converter = self._mills_converter
        else:  # TimestampNtz
            converter = self._mills_converter_utc

        microsec = value.microsecond
        nanosec = value.nanosecond
        secs = int(converter.to_milliseconds(value.to_pydatetime()) / 1000)
        value = pd.Timestamp(converter.from_milliseconds(secs * 1000))
        return value.replace(microsecond=microsec, nanosecond=nanosec)

    def _convert_struct_timestamp(self, value, ntz=False):
        if value is None:
            return None

        if not ntz:
            converter = self._mills_converter
        else:  # TimestampNtz
            converter = self._mills_converter_utc

        ts = pd.Timestamp(converter.from_milliseconds(value["sec"] * 1000))
        nanos = value["nano"]
        return ts.replace(microsecond=nanos // 1000, nanosecond=nanos % 1000)

    @staticmethod
    def _convert_struct_timedelta(value):
        if value is None:
            return None

        nanos = value["nano"]
        return pd.Timedelta(
            seconds=value["sec"], microseconds=nanos // 1000, nanoseconds=nanos % 1000
        )

    @staticmethod
    def _convert_struct(value, field_converters, tuple_type):
        if value is None:
            return None

        results = {
            k: field_converters[k](v) for k, v in value.items()
        }
        if tuple_type is not None:
            return tuple_type(**results)
        else:
            return results

    def _build_converter(self, odps_type, arrow_type=None):
        import pyarrow as pa

        arrow_decimal_types = (pa.Decimal128Type,)
        if hasattr(pa, "Decimal256Type"):
            arrow_decimal_types += (pa.Decimal256Type,)

        if self._convert_ts and isinstance(odps_type, types.Datetime):
            return self._convert_datetime
        elif isinstance(odps_type, types.Timestamp):
            if isinstance(arrow_type, pa.StructType):
                return self._convert_struct_timestamp
            elif self._convert_ts:
                return self._convert_ts_timestamp
            else:
                return _reflective
        elif isinstance(odps_type, types.TimestampNTZ):
            if isinstance(arrow_type, pa.StructType):
                return functools.partial(self._convert_struct_timestamp, ntz=True)
            elif self._convert_ts:
                return functools.partial(self._convert_ts_timestamp, ntz=True)
            else:
                return _reflective
        elif (
            isinstance(odps_type, types.Decimal)
            and isinstance(arrow_type, pa.FixedSizeBinaryType)
            and not isinstance(arrow_type, arrow_decimal_types)
        ):
            return convert_legacy_decimal_bytes
        elif (
            isinstance(odps_type, types.IntervalDayTime)
            and isinstance(arrow_type, pa.StructType)
        ):
            return self._convert_struct_timedelta
        elif isinstance(odps_type, types.Array):
            arrow_value_type = getattr(arrow_type, "value_type", None)
            sub_converter = self._build_converter(odps_type.value_type, arrow_value_type)
            if sub_converter is _reflective:
                return _reflective
            return lambda value: [sub_converter(x) for x in value]
        elif isinstance(odps_type, types.Map):
            arrow_key_type = getattr(arrow_type, "key_type", None)
            arrow_value_type = getattr(arrow_type, "item_type", None)

            key_converter = self._build_converter(odps_type.key_type, arrow_key_type)
            value_converter = self._build_converter(odps_type.value_type, arrow_value_type)
            if key_converter is _reflective and value_converter is _reflective:
                return OrderedDict

            return lambda value: OrderedDict(
                [(key_converter(k), value_converter(v)) for k, v in value]
            )
        elif isinstance(odps_type, types.Struct):
            field_converters = {}
            for field_name, field_type in odps_type.field_types.items():
                arrow_field_type = None
                if arrow_type is not None:
                    arrow_field_type = arrow_type[field_name].type
                field_converters[field_name] = self._build_converter(
                    field_type, arrow_field_type
                )

            if options.struct_as_dict:
                tuple_type = None
            else:
                tuple_type = odps_type.namedtuple_type
            return functools.partial(
                self._convert_struct, field_converters=field_converters, tuple_type=tuple_type
            )
        else:
            return _reflective

    def __call__(self, value):
        return self._converter(value)


class ArrowRecordReader(AbstractRecordReader):
    _complex_types_to_convert = (types.Array, types.Map, types.Struct)

    def __init__(self, arrow_reader, make_compat=True, read_all=False):
        self._arrow_reader = arrow_reader
        self._batch_pos = 0
        self._total_pos = 0
        self._cur_batch = None
        self._make_compat = make_compat

        self._field_converters = None

        if read_all:
            self._cur_batch = arrow_reader.read()

    def _convert_record(self, arrow_values):
        py_values = [x.as_py() for x in arrow_values]
        if not self._make_compat:
            return py_values
        else:
            return [
                converter(value)
                for value, converter in zip(py_values, self._field_converters)
            ]

    def read(self):
        if self._cur_batch is None or self._batch_pos >= self._cur_batch.num_rows:
            self._cur_batch = self._arrow_reader.read_next_batch()
            self._batch_pos = 0

            if self._cur_batch is None or self._cur_batch.num_rows == 0:
                return None

        if self._field_converters is None:
            table_schema = self._arrow_reader.schema
            self._field_converters = [
                ArrowRecordFieldConverter(table_schema[col_name].type, arrow_type)
                for col_name, arrow_type in zip(
                    self._cur_batch.schema.names, self._cur_batch.schema.types
                )
            ]
        tp = tuple(col[self._batch_pos] for col in self._cur_batch.columns)
        self._batch_pos += 1
        self._total_pos += 1

        return Record(schema=self.schema, values=self._convert_record(tp))

    def to_pandas(self, start=None, count=None, **kw):
        step = kw.get("step") or 1
        return self._arrow_reader.to_pandas().iloc[start: start + count: step]

    def close(self):
        self._arrow_reader.close()

    def __next__(self):
        rec = self.read()
        if rec is None:
            raise StopIteration
        return rec

    next = __next__

    @property
    def count(self):
        return self._total_pos

    @property
    def schema(self):
        return self._arrow_reader.schema

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from collections import namedtuple

from ... import options, compat
from . import _customize
from .utils import get_field_name_by_role, detect_metrics_fallback, metrics_result

try:
    import numpy as np
except ImportError:
    from warnings import warn
    warn('Numpy not installed. Metrics module will not be available.')
    np = None

from ..enums import FieldRole

logger = logging.getLogger(__name__)


def _run_cm_node(df, col_true, col_pred, execute_now=True, result_callback=None):
    detect_metrics_fallback(df)
    if not options.ml.use_old_metrics:
        if result_callback:
            cb = lambda v: result_callback(v.confusion_matrix)
        else:
            cb = lambda v: (v.confusion_matrix, v.labels)

        eval_fun = getattr(_customize, 'eval_multi_class')
        eval_result = eval_fun(df, label_col=col_true, predict_col=col_pred,
                               execute_now=execute_now, _result_callback=cb)
        return eval_result
    else:
        cb = result_callback or (lambda v: v)
        cm_fun = getattr(_customize, '_confusion_matrix')
        return cm_fun(df, label_col=col_true, predict_col=col_pred,
                      execute_now=execute_now, _result_callback=cb)


def _run_roc_node(df, pos_label, col_true, col_pred, col_scores, execute_now=True, result_callback=None):
    detect_metrics_fallback(df)
    RocResult = _customize.RocResult

    if not options.ml.use_old_metrics:
        if result_callback:
            cb = lambda v: result_callback(RocResult(v.thresh, v.tp, v.fn, v.tn, v.fp))
        else:
            cb = lambda v: RocResult(v.thresh, v.tp, v.fn, v.tn, v.fp)

        eval_fun = getattr(_customize, 'eval_binary_class')
        eval_result = eval_fun(df, good_value=pos_label, label_col=col_true, score_col=col_scores,
                               execute_now=execute_now, _result_callback=cb)
        return eval_result
    else:
        result_callback = result_callback or (lambda v: v)
        roc_fun = getattr(_customize, '_roc')
        return roc_fun(df, good_value=pos_label, label_col=col_true, predict_col=col_pred,
                       score_col=col_scores, execute_now=execute_now, _result_callback=result_callback)


@metrics_result(_run_cm_node)
def confusion_matrix(df, col_true=None, col_pred=None):
    """
    Compute confusion matrix of a predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param col_true: column name of true label
    :type col_true: str
    :param col_true: column name of predicted label, 'prediction_result' by default.
    :type col_pred: str
    :return: Confusion matrix and mapping list for classes

    :Example:

    >>> predicted = model.predict(input_data)
    >>> cm, mapping = confusion_matrix(predicted, 'category')
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_CLASS)
    return _run_cm_node(df, col_true, col_pred)[0]


@metrics_result(_run_cm_node)
def accuracy_score(df, col_true=None, col_pred=None, normalize=True):
    """
    Compute accuracy of a predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param col_true: column name of true label
    :type col_true: str
    :param col_true: column name of predicted label, 'prediction_result' by default.
    :type col_pred: str
    :param normalize: denoting if the output is normalized between [0, 1]
    :type normalize: bool
    :return: Accuracy value
    :rtype: float
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_CLASS)
    mat, _ = _run_cm_node(df, col_true, col_pred)
    if np is not None:
        acc_count = np.sum(np.diag(mat))
        if not normalize:
            return acc_count
        else:
            return acc_count * 1.0 / np.sum(mat)
    else:
        diag_sum = mat_sum = 0
        mat_size = len(mat)
        for i in compat.irange(mat_size):
            for j in compat.irange(mat_size):
                if i == j:
                    diag_sum += mat[i][j]
                mat_sum += mat[i][j]
        if not normalize:
            return diag_sum
        else:
            return diag_sum * 1.0 / mat_sum


@metrics_result(_run_cm_node)
def precision_score(df, col_true=None, col_pred='precision_result', pos_label=1, average=None):
    r"""
    Compute precision of a predicted DataFrame. Precision is defined as :math:`\frac{TP}{TP + TN}`

    :Parameters:
        - **df** - predicted data frame
        - **col_true** - column name of true label
        - **col_pred** - column name of predicted label, 'prediction_result' by default.
        - **pos_label** - denote the desired class label when ``average`` == `binary`
        - **average** - denote the method to compute average.
    :Returns:
        Precision score
    :Return type:
        float or numpy.array[float]

    The parameter ``average`` controls the behavior of the function.

    - When ``average`` == None (by default), precision of every class is given as a list.

    - When ``average`` == 'binary', precision of class specified in ``pos_label`` is given.

    - When ``average`` == 'micro', STP / (STP + STN) is given, where STP and STN are summations of TP and TN for every class.

    - When ``average`` == 'macro', average precision of all the class is given.

    - When ``average`` == `weighted`, average precision of all the class weighted by support of every true classes is given.

    :Example:

    Assume we have a table named 'predicted' as follows:

    ======== ===================
    label    prediction_result
    ======== ===================
    0        0
    1        2
    2        1
    0        0
    1        0
    2        1
    ======== ===================

    Different options of ``average`` parameter outputs different values:

.. code-block:: python

    >>> precision_score(predicted, 'label', average=None)
    array([ 0.66...,  0.        ,  0.        ])
    >>> precision_score(predicted, 'label', average='macro')
    0.22
    >>> precision_score(predicted, 'label', average='micro')
    0.33
    >>> precision_score(predicted, 'label', average='weighted')
    0.22
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_CLASS)
    mat, label_list = _run_cm_node(df, col_true, col_pred)
    class_dict = dict((label, idx) for idx, label in enumerate(label_list))
    tps = np.diag(mat)
    pred_count = np.sum(mat, axis=0)
    if average is None:
        return tps * 1.0 / pred_count
    elif average == 'binary':
        class_idx = class_dict[pos_label]
        return tps[class_idx] * 1.0 / pred_count[class_idx]
    elif average == 'micro':
        return np.sum(tps) / np.sum(pred_count)
    elif average == 'macro':
        return np.mean(tps * 1.0 / pred_count)
    elif average == 'weighted':
        support = np.sum(mat, axis=1)
        return np.sum(tps * 1.0 / pred_count * support) / np.sum(support)


@metrics_result(_run_cm_node)
def recall_score(df, col_true=None, col_pred='precision_result', pos_label=1, average=None):
    r"""
    Compute recall of a predicted DataFrame. Precision is defined as :math:`\frac{TP}{TP + FP}`

    :Parameters:
        - **df* - predicted data frame
        - **col_true** - column name of true label
        - **col_pred** - column name of predicted label, 'prediction_result' by default.
        - **pos_label** - denote the desired class label when ``average`` == `binary`
        - **average** - denote the method to compute average.
    :Returns:
        Recall score
    :Return type:
        float | numpy.array[float]

    The parameter ``average`` controls the behavior of the function.

    * When ``average`` == None (by default), recall of every class is given as a list.

    * When ``average`` == 'binary', recall of class specified in ``pos_label`` is given.

    * When ``average`` == 'micro', STP / (STP + SFP) is given, where STP and SFP are summations of TP and FP for every class.

    * When ``average`` == 'macro', average recall of all the class is given.

    * When ``average`` == `weighted`, average recall of all the class weighted by support of every true classes is given.

    :Example:

    Assume we have a table named 'predicted' as follows:

    ======== ===================
    label    prediction_result
    ======== ===================
    0        1
    1        2
    2        1
    1        1
    1        0
    2        2
    ======== ===================

    Different options of ``average`` parameter outputs different values:

.. code-block:: python

    >>> recall_score(predicted, 'label', average=None)
    array([ 0.        ,  0.33333333,  0.5       ])
    >>> recall_score(predicted, 'label', average='macro')
    0.27
    >>> recall_score(predicted, 'label', average='micro')
    0.33
    >>> recall_score(predicted, 'label', average='weighted')
    0.33
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_CLASS)
    mat, label_list = _run_cm_node(df, col_true, col_pred)
    class_dict = dict((label, idx) for idx, label in enumerate(label_list))
    tps = np.diag(mat)
    supp_count = np.sum(mat, axis=1)
    if average is None:
        return tps * 1.0 / supp_count
    elif average == 'binary':
        class_idx = class_dict[pos_label]
        return tps[class_idx] * 1.0 / supp_count[class_idx]
    elif average == 'micro' or average == 'weighted':
        return np.sum(tps) / np.sum(supp_count)
    elif average == 'macro':
        return np.mean(tps * 1.0 / supp_count)


@metrics_result(_run_cm_node)
def fbeta_score(df, col_true=None, col_pred='precision_result', beta=1.0, pos_label=1, average=None):
    r"""
    Compute f-beta score of a predicted DataFrame. f-beta is defined as

    .. math::

        \frac{1 + \beta^2 \cdot precision \cdot recall}{\beta^2 \cdot precision + recall}

    F-beta score is a generalization of f-1 score.

    :Parameters:
        - **df** - predicted data frame
        - **col_true** - column name of true label
        - **col_pred** - column name of predicted label, 'prediction_result' by default.
        - **pos_label** - denote the desired class label when ``average`` == `binary`
        - **average** - denote the method to compute average.
    :Returns:
        Recall score
    :Return type:
        float | numpy.array[float]

    The parameter ``average`` controls the behavior of the function.

    * When ``average`` == None (by default), f-beta of every class is given as a list.

    * When ``average`` == 'binary', f-beta of class specified in ``pos_label`` is given.

    * When ``average`` == 'micro', f-beta of overall precision and recall is given, where overall precision and recall are computed in micro-average mode.

    * When ``average`` == 'macro', average f-beta of all the class is given.

    * When ``average`` == `weighted`, average f-beta of all the class weighted by support of every true classes is given.

    :Example:

    Assume we have a table named 'predicted' as follows:

    ======== ===================
    label    prediction_result
    ======== ===================
    0        1
    1        2
    2        1
    1        1
    1        0
    2        2
    ======== ===================

    Different options of ``average`` parameter outputs different values:

.. code-block:: python

    >>> fbeta_score(predicted, 'label', average=None, beta=0.5)
    array([ 0.        ,  0.33333333,  0.5       ])
    >>> fbeta_score(predicted, 'label', average='macro', beta=0.5)
    0.27
    >>> fbeta_score(predicted, 'label', average='micro', beta=0.5)
    0.33
    >>> fbeta_score(predicted, 'label', average='weighted', beta=0.5)
    0.33
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_CLASS)
    mat, label_list = _run_cm_node(df, col_true, col_pred)
    class_dict = dict((label, idx) for idx, label in enumerate(label_list))
    tps = np.diag(mat)
    pred_count = np.sum(mat, axis=0)
    supp_count = np.sum(mat, axis=1)
    beta2 = beta ** 2

    precision = tps * 1.0 / pred_count
    recall = tps * 1.0 / supp_count
    ppr = precision * beta2 + recall
    ppr[ppr == 0] = 1e-6

    fbeta = (1 + beta2) * precision * recall / ppr

    if average is None:
        return fbeta
    elif average == 'binary':
        class_idx = class_dict[pos_label]
        return fbeta[class_idx]
    elif average == 'micro':
        g_precision = np.sum(tps) * 1.0 / np.sum(supp_count)
        g_recall = np.sum(tps) * 1.0 / np.sum(pred_count)
        return (1 + beta2) * g_precision * g_recall / (beta2 * g_precision + g_recall)
    elif average == 'macro':
        return np.mean(fbeta)
    elif average == 'weighted':
        return sum(fbeta * supp_count) / sum(supp_count)


@metrics_result(_run_roc_node)
def f1_score(df, col_true=None, col_pred='precision_result', pos_label=1, average=None):
    r"""
    Compute f-1 score of a predicted DataFrame. f-1 is defined as

    .. math::

        \frac{2 \cdot precision \cdot recall}{precision + recall}


    :Parameters:
        - **df** - predicted data frame
        - **col_true** - column name of true label
        - **col_pred** - column name of predicted label, 'prediction_result' by default.
        - **pos_label** - denote the desired class label when ``average`` == `binary`
        - **average** - denote the method to compute average.
    :Returns:
        Recall score
    :Return type:
        float | numpy.array[float]

    The parameter ``average`` controls the behavior of the function.

    * When ``average`` == None (by default), f-1 of every class is given as a list.

    * When ``average`` == 'binary', f-1 of class specified in ``pos_label`` is given.

    * When ``average`` == 'micro', f-1 of overall precision and recall is given, where overall precision and recall are computed in micro-average mode.

    * When ``average`` == 'macro', average f-1 of all the class is given.

    * When ``average`` == `weighted`, average f-1 of all the class weighted by support of every true classes is given.


    :Example:

    Assume we have a table named 'predicted' as follows:

    ======== ===================
    label    prediction_result
    ======== ===================
    0        1
    1        2
    2        1
    1        1
    1        0
    2        2
    ======== ===================

    Different options of ``average`` parameter outputs different values:

.. code-block:: python

    >>> f1_score(predicted, 'label', average=None)
    array([ 0.        ,  0.33333333,  0.5       ])
    >>> f1_score(predicted, 'label', average='macro')
    0.27
    >>> f1_score(predicted, 'label', average='micro')
    0.33
    >>> f1_score(predicted, 'label', average='weighted')
    0.33
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_CLASS)
    return fbeta_score(df, col_true, col_pred, pos_label=pos_label, average=average)


@metrics_result(_run_roc_node)
def roc_curve(df, col_true=None, col_pred=None, col_scores=None, pos_label=1):
    r"""
    Compute true positive rate (TPR), false positive rate (FPR) and threshold from predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param pos_label: positive label
    :type pos_label: str
    :param col_true: true column
    :type col_true: str
    :param col_pred: predicted column, 'prediction_result' if absent.
    :type col_pred: str
    :param col_scores: score column, 'prediction_score' if absent.
    :type col_scores: str

    :return: False positive rate, true positive rate and threshold, in numpy array format.

    :Example:

    >>> import matplotlib.pyplot as plt
    >>> fpr, tpr, thresh = roc_curve(predicted, "class")
    >>> plt.plot(fpr, tpr)
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_CLASS)
    if not col_scores:
        col_scores = get_field_name_by_role(df, FieldRole.PREDICTED_SCORE)
    thresh, tp, fn, tn, fp = _run_roc_node(df, pos_label, col_true, col_pred, col_scores)

    if np is not None:
        tpr = tp * 1.0 / (tp + fn)
        fpr = fp * 1.0 / (fp + tn)
    else:
        tpr = [tp[i] * 1.0 / (tp[i] + fn[i]) for i in range(len(tp))]
        fpr = [fp[i] * 1.0 / (fp[i] + tn[i]) for i in range(len(fp))]

    roc_result = namedtuple('ROCResult', 'fpr tpr thresh')
    return roc_result(fpr=fpr, tpr=tpr, thresh=thresh)


@metrics_result(_run_roc_node)
def gain_chart(df, col_true=None, col_pred=None, col_scores=None, pos_label=1):
    r"""
    Compute positive proportion, true positive rate (TPR) and threshold from predicted DataFrame. The trace can be plotted as a cumulative gain chart

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param pos_label: positive label
    :type pos_label: str
    :param col_true: true column
    :type col_true: str
    :param col_pred: predicted column, 'prediction_result' if absent.
    :type col_pred: str
    :param col_scores: score column, 'prediction_score' if absent.
    :type col_scores: str

    :return: positive proportion, true positive rate and threshold, in numpy array format.

    :Example:

    >>> import matplotlib.pyplot as plt
    >>> depth, tpr, thresh = gain_chart(predicted)
    >>> plt.plot(depth, tpr)
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_CLASS)
    if not col_scores:
        col_scores = get_field_name_by_role(df, FieldRole.PREDICTED_SCORE)
    thresh, tp, fn, tn, fp = _run_roc_node(df, pos_label, col_true, col_pred, col_scores)

    depth = (tp + fp) * 1.0 / (tp + fp + tn + fn)
    tpr = tp * 1.0 / (tp + fn)

    gain_result = namedtuple('GainChartResult', 'depth tpr thresh')
    return gain_result(depth=depth, tpr=tpr, thresh=thresh)


@metrics_result(_run_roc_node)
def lift_chart(df, col_true=None, col_pred=None, col_scores=None, pos_label=1):
    r"""
    Compute life value, true positive rate (TPR) and threshold from predicted DataFrame.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param pos_label: positive label
    :type pos_label: str
    :param col_true: true column
    :type col_true: str
    :param col_pred: predicted column, 'prediction_result' if absent.
    :type col_pred: str
    :param col_scores: score column, 'prediction_score' if absent.
    :type col_scores: str

    :return: lift value, true positive rate and threshold, in numpy array format.

    :Example:

    >>> import matplotlib.pyplot as plt
    >>> depth, lift, thresh = lift_chart(predicted)
    >>> plt.plot(depth, lift)
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_CLASS)
    if not col_scores:
        col_scores = get_field_name_by_role(df, FieldRole.PREDICTED_SCORE)
    thresh, tp, fn, tn, fp = _run_roc_node(df, pos_label, col_true, col_pred, col_scores)

    depth = (tp + fp) * 1.0 / (tp + fp + tn + fn)
    tpr = tp * 1.0 / (tp + fn)
    lift = tpr / depth

    lift_result = namedtuple('LiftResult', 'depth lift thresh')
    return lift_result(depth=depth, lift=lift, thresh=thresh)


def auc(tpr, fpr):
    """
    Calculate AUC value from true positive rate (TPR) and false positive rate (FPR)\
    with trapezoidal rule.

    Note that calculation on DataFrames should use ``roc_auc_score`` instead.

    :param tpr: True positive rate array
    :param fpr: False positive rate array
    :return: AUC value
    :rtype: float
    """
    return abs(np.trapz(tpr, fpr))


@metrics_result(_run_roc_node)
def roc_auc_score(df, col_true=None, col_pred=None, col_scores=None, pos_label=1):
    """
    Compute Area Under the Curve (AUC) from prediction scores with trapezoidal rule.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param pos_label: positive label
    :type pos_label: str
    :param col_true: true column
    :type col_true: str
    :param col_pred: predicted column, 'prediction_result' if absent.
    :type col_pred: str
    :param col_scores: score column, 'prediction_score' if absent.
    :type col_scores: str
    :return: AUC value
    :rtype: float
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_CLASS)
    if not col_scores:
        col_scores = get_field_name_by_role(df, FieldRole.PREDICTED_SCORE)
    thresh, tp, fn, tn, fp = _run_roc_node(df, pos_label, col_true, col_pred, col_scores)
    tpr = tp * 1.0 / (tp + fn)
    fpr = fp * 1.0 / (fp + tn)
    return auc(tpr, fpr)


@metrics_result(_run_roc_node)
def precision_recall_curve(df, col_true=None, col_pred=None, col_scores=None, pos_label=1):
    """
    Compute precision and recall value with different thresholds. These precision and recall\
      values can be used to plot a precision-recall curve.

    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param pos_label: positive label
    :type pos_label: str
    :param col_true: true column
    :type col_true: str
    :param col_pred: predicted column, 'prediction_result' if absent.
    :type col_pred: str
    :param col_scores: score column, 'prediction_score' if absent.
    :type col_scores: str
    :return: precision, recall and threshold, in numpy arrays.
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_CLASS)
    if not col_scores:
        col_scores = get_field_name_by_role(df, FieldRole.PREDICTED_SCORE)
    thresh, tp, fn, tn, fp = _run_roc_node(df, pos_label, col_true, col_pred, col_scores)

    precisions = np.squeeze(np.asarray(tp * 1.0 / (tp + fp)))
    recalls = np.squeeze(np.asarray(tp * 1.0 / (tp + fn)))

    result_type = namedtuple('PrecisionRecallResult', 'precisions recalls thresh')
    return result_type(precisions=precisions, recalls=recalls, thresh=thresh)


@metrics_result(_run_roc_node)
def average_precision_score(df, col_true=None, col_pred=None, col_scores=None, pos_label=1):
    """
    Compute average precision score, i.e., the area under precision-recall curve.
    
    Note that this method will trigger the defined flow to execute.

    :param df: predicted data frame
    :type df: DataFrame
    :param pos_label: positive label
    :type pos_label: str
    :param col_true: true column
    :type col_true: str
    :param col_pred: predicted column, 'prediction_result' if absent.
    :type col_pred: str
    :param col_scores: score column, 'prediction_score' if absent.
    :type col_scores: str
    :return: Average precision score
    :rtype: float
    """
    if not col_pred:
        col_pred = get_field_name_by_role(df, FieldRole.PREDICTED_CLASS)
    if not col_scores:
        col_scores = get_field_name_by_role(df, FieldRole.PREDICTED_SCORE)
    thresh, tp, fn, tn, fp = _run_roc_node(df, pos_label, col_true, col_pred, col_scores)

    precisions = np.squeeze(np.asarray(tp * 1.0 / (tp + fp)))
    recalls = np.squeeze(np.asarray(tp * 1.0 / (tp + fn)))

    return np.trapz(precisions, recalls)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import tarfile
import zipfile

from .... import compat, options
from ....compat import six
from ....errors import ODPSError
from ....lib.importer import CompressImporter
from ....models import Table, TableSchema, Partition as TableSchemaPartition
from ....models.partition import Partition
from ... import DataFrame
from ...expr.core import ExprDictionary
from ...expr.expressions import CollectionExpr, Scalar
from ...expr.dynamic import DynamicMixin
from ...backends.odpssql.types import df_schema_to_odps_schema, df_type_to_odps_type
from ...utils import is_source_collection, is_constant_scalar
from ...types import DynamicSchema, Unknown
from ..context import context
from ..core import Engine, ExecuteNode, ExprDAG
from ..errors import CompileError
from ..frame import ResultFrame
from ..utils import refresh_dynamic, write_table
from . import analyzer as ana
from .compiler import PandasCompiler
from .types import pd_to_df_schema


class PandasExecuteNode(ExecuteNode):
    def __repr__(self):
        return 'Local execution by pandas backend'

    def _repr_html_(self):
        return '<p>Local execution by pandas backend</p>'


def with_thirdparty_libs(fun):
    def wrapped(self, *args, **kwargs):
        libraries = self._get_libraries(kwargs.get('libraries'))
        importer = self._build_library_importer(libraries)
        if importer is not None:
            sys.meta_path.append(importer)

        try:
            return fun(self, *args, **kwargs)
        finally:
            if importer is not None:
                sys.meta_path = [p for p in sys.meta_path if p is not importer]

    wrapped.__name__ = fun.__name__
    wrapped.__doc__ = fun.__doc__
    return wrapped


class PandasEngine(Engine):
    def __init__(self, odps=None):
        self._odps = odps
        self._file_objs = []

    def _new_execute_node(self, expr_dag):
        return PandasExecuteNode(expr_dag)

    def _run(self, expr_dag, pd_dag, ui=None, progress_proportion=1, **_):
        ui.status('Try to execute by local pandas...', clear_keys=True)

        results = ExprDictionary()
        while True:
            topos = pd_dag.topological_sort()

            no_sub = True
            for node in topos:
                expr, func = node
                if expr in results:
                    continue
                res = func(results)
                if isinstance(res, tuple):
                    src = expr
                    expr = res[0]
                    res = res[1]
                    results[src] = res
                    results[expr] = res

                    # break cuz the dag has changed
                    no_sub = False
                    break
                results[expr] = res
            if no_sub:
                break

        ui.inc(progress_proportion)

        try:
            return results[expr_dag.root]
        except KeyError as e:
            if len(results) == 1:
                return compat.lvalues(results)[0]
            raise e
        finally:
            for fo in self._file_objs:
                try:
                    fo.close()
                except:
                    pass
            self._file_objs = []

    def _new_analyzer(self, expr_dag, on_sub=None):
        return ana.Analyzer(expr_dag)

    def _compile(self, expr_dag):
        backend = PandasCompiler(expr_dag)
        return backend.compile(expr_dag.root)

    def _cache(self, expr_dag, dag, expr, **kwargs):
        import pandas as pd

        if is_source_collection(expr_dag.root) or \
                is_constant_scalar(expr_dag.root):
            return

        execute_dag = ExprDAG(expr_dag.root, dag=expr_dag)

        if isinstance(expr, CollectionExpr):
            root = expr_dag.root
            sub = CollectionExpr(_source_data=pd.DataFrame(), _schema=expr.schema)
            sub.add_deps(root)
            expr_dag.substitute(root, sub)

            execute_node = self._execute(execute_dag, dag, expr, ret_df=True, **kwargs)

            def callback(res):
                for col in res.columns:
                    sub._source_data[col] = res[col]
                if isinstance(expr, DynamicMixin):
                    sub._schema = pd_to_df_schema(res)
                    refresh_dynamic(sub, expr_dag)

            execute_node.callback = callback
        else:
            assert isinstance(expr, Scalar)  # sequence is not cache-able

            class ValueHolder(object): pass
            sub = Scalar(_value_type=expr.dtype)
            sub._value = ValueHolder()
            root = expr_dag.root
            sub.add_deps(root)
            expr_dag.substitute(root, sub)

            execute_node = self._execute(execute_dag, dag, expr, **kwargs)

            def callback(res):
                sub._value = res

            execute_node.callback = callback

        return sub, execute_node

    def _build_library_importer(self, libraries):
        if libraries is None:
            return None

        def _open_file(*args, **kwargs):
            handle = open(*args, **kwargs)
            self._file_objs.append(handle)
            return handle

        readers = []

        for lib in libraries:
            if isinstance(lib, six.string_types):
                lib = os.path.abspath(lib)
                file_dict = dict()
                if os.path.isfile(lib):
                    file_dict[lib.replace(os.path.sep, '/')] = _open_file(lib, 'rb')
                else:
                    for root, dirs, files in os.walk(lib):
                        for f in files:
                            fpath = os.path.join(root, f)
                            file_dict[fpath.replace(os.path.sep, '/')] = _open_file(fpath, 'rb')
                readers.append(file_dict)
            else:
                lib_name = lib.name
                if lib_name.endswith('.zip') or lib_name.endswith('.egg') or lib_name.endswith('.whl'):
                    readers.append(zipfile.ZipFile(lib.open(mode='rb')))
                elif lib_name.endswith('.tar') or lib_name.endswith('.tar.gz') or lib_name.endswith('.tar.bz2'):
                    if lib_name.endswith('.tar'):
                        mode = 'r'
                    else:
                        mode = 'r:gz' if lib_name.endswith('.tar.gz') else 'r:bz2'
                    readers.append(tarfile.open(fileobj=six.BytesIO(lib.open(mode='rb').read()), mode=mode))
                elif lib_name.endswith('.py'):
                    tarbinary = six.BytesIO()
                    tar = tarfile.open(fileobj=tarbinary, mode='w:gz')
                    fbin = lib.open(mode='rb').read()
                    info = tarfile.TarInfo(name='pyodps_lib/' + lib_name)
                    info.size = len(fbin)
                    tar.addfile(info, fileobj=six.BytesIO(fbin))
                    tar.close()
                    readers.append(tarfile.open(fileobj=six.BytesIO(tarbinary.getvalue()), mode='r:gz'))
                else:
                    raise ValueError(
                        'Unknown library type which should be one of zip(egg, wheel), tar, or tar.gz')

        return CompressImporter(*readers, extract_binary=True, supersede=options.df.supersede_libraries)

    @with_thirdparty_libs
    def _do_execute(self, expr_dag, expr, ui=None, progress_proportion=1,
                    head=None, tail=None, **kw):
        expr_dag = self._convert_table(expr_dag)
        self._rewrite(expr_dag)

        ret_df = kw.pop('ret_df', False)

        src_expr = expr

        pd_dag = self._compile(expr_dag)
        df = self._run(expr_dag, pd_dag, ui=ui, progress_proportion=progress_proportion,
                       **kw)

        if not isinstance(src_expr, Scalar):
            need_cache = False
            if not isinstance(src_expr, CollectionExpr) or getattr(src_expr, "_source_data", None) is None:
                # only cache when expression is not data source
                need_cache = True
                context.cache(src_expr, df)
            # reset schema
            if isinstance(src_expr, CollectionExpr) and \
                    (isinstance(src_expr._schema, DynamicSchema) or
                         any(isinstance(col.type, Unknown) for col in src_expr._schema.columns)):
                src_expr._schema = expr_dag.root.schema
            if head:
                df = df[:head]
            elif tail:
                df = df[-tail:]
            if ret_df:
                return df

            if need_cache:
                df = df.copy()
            return ResultFrame(df, schema=expr_dag.root.schema)
        else:
            res = df.values[0][0]
            context.cache(src_expr, res)
            return res

    @with_thirdparty_libs
    def _do_persist(self, expr_dag, expr, name, ui=None, project=None,
                    partitions=None, partition=None, odps=None, lifecycle=None,
                    progress_proportion=1, execute_percent=0.5, overwrite=True,
                    drop_table=False, create_table=True, drop_partition=False,
                    create_partition=False, cast=False, schema=None, **kwargs):
        expr_dag = self._convert_table(expr_dag)
        self._rewrite(expr_dag)

        if isinstance(name, Partition):
            partition = name.partition_spec
            name = name.table
        if isinstance(name, Table):
            table = name
            project = table.project.name
            if table.get_schema():
                schema = table.get_schema().name
            name = table.name

        src_expr = expr
        expr = expr_dag.root
        odps = odps or self._odps
        if odps is None:
            raise ODPSError('ODPS entrance should be provided')
        schema = schema or odps.schema

        df = self._do_execute(expr_dag, src_expr, ui=ui,
                              progress_proportion=progress_proportion * execute_percent, **kwargs)
        t_schema = TableSchema(columns=df.columns)

        if drop_table:
            odps.delete_table(name, project=project, schema=schema, if_exists=True)

        if partitions is not None:
            if drop_partition:
                raise ValueError('Cannot drop partitions when specify `partitions`')
            if create_partition:
                raise ValueError('Cannot create partitions when specify `partitions`')

            if isinstance(partitions, tuple):
                partitions = list(partitions)
            if not isinstance(partitions, list):
                partitions = [partitions, ]

            for p in partitions:
                if p not in t_schema:
                    raise ValueError(
                            'Partition field(%s) does not exist in DataFrame schema' % p)

            t_schema = df_schema_to_odps_schema(t_schema)
            columns = [c for c in t_schema.columns if c.name not in partitions]
            ps = [TableSchemaPartition(name=t, type=t_schema.get_type(t)) for t in partitions]
            t_schema = TableSchema(columns=columns, partitions=ps)

            if odps.exist_table(name, project=project, schema=schema) or not create_table:
                t = odps.get_table(name, project=project, schema=schema)
            else:
                t = odps.create_table(name, t_schema, project=project, schema=schema, lifecycle=lifecycle)
        elif partition is not None:
            if odps.exist_table(name, project=project, schema=schema) or not create_table:
                t = odps.get_table(name, project=project, schema=schema)
                partition = self._get_partition(partition, t)

                if drop_partition:
                    t.delete_partition(partition, if_exists=True)
                if create_partition:
                    t.create_partition(partition, if_not_exists=True)
            else:
                partition = self._get_partition(partition)
                project_obj = odps.get_project(project)
                column_names = [n for n in expr.schema.names if n not in partition]
                column_types = [
                    df_type_to_odps_type(expr.schema[n].type, project=project_obj)
                    for n in column_names
                ]
                partition_names = [n for n in partition.keys]
                partition_types = ['string'] * len(partition_names)
                t = odps.create_table(
                    name, TableSchema.from_lists(
                        column_names, column_types, partition_names, partition_types
                    ),
                    project=project, lifecycle=lifecycle, schema=schema
                )
                if create_partition is None or create_partition is True:
                    t.create_partition(partition)
        else:
            if drop_partition:
                raise ValueError('Cannot drop partition for non-partition table')
            if create_partition:
                raise ValueError('Cannot create partition for non-partition table')

            if odps.exist_table(name, project=project, schema=schema) or not create_table:
                t = odps.get_table(name, project=project, schema=schema)
                if t.table_schema.partitions:
                    raise CompileError('Cannot insert into partition table %s without specifying '
                                       '`partition` or `partitions`.')
            else:
                t = odps.create_table(
                    name,
                    df_schema_to_odps_schema(t_schema),
                    project=project,
                    lifecycle=lifecycle,
                    schema=schema,
                )

        write_table(df, t, ui=ui, cast=cast, overwrite=overwrite, partitions=partitions, partition=partition,
                    progress_proportion=progress_proportion*(1-execute_percent))

        if partition:
            if partition:
                filters = []
                df = DataFrame(t)
                for k in partition.keys:
                    filters.append(df[k] == partition[k])
                return df.filter(*filters)
        return DataFrame(t)
